"""Configuration file for examples."""
import matplotlib.pyplot as plt
from IPython.display import set_matplotlib_formats

set_matplotlib_formats("pdf", "png")
plt.rcParams["savefig.dpi"] = 75
plt.rcParams["figure.autolayout"] = False
plt.rcParams["figure.figsize"] = 10, 6



"""LOF test."""
import os

import numpy as np
import pytest
import scipy.io as sio

from meegkit.lof import LOF

rng = np.random.default_rng(10)

# Data files
THIS_FOLDER = os.path.dirname(os.path.abspath(__file__)) # data folder of MEEGKIT


@pytest.mark.parametrize(argnames="n_neighbors", argvalues=(8, 20, 40, 2048))
def test_lof(n_neighbors, show=False):
    mat = sio.loadmat(os.path.join(THIS_FOLDER, "data", "lofdata.mat"))
    X = mat["X"]
    lof = LOF(n_neighbors)
    bad_channel_indices = lof.predict(X)
    print(bad_channel_indices)

@pytest.mark.parametrize(argnames="metric",
                         argvalues=("euclidean", "nan_euclidean",
                                    "cosine", "cityblock", "manhattan"))
def test_lof2(metric, show=False):
    mat = sio.loadmat(os.path.join(THIS_FOLDER, "data", "lofdata.mat"))
    X = mat["X"]
    lof = LOF(20, metric)
    bad_channel_indices = lof.predict(X)
    print(bad_channel_indices)

if __name__ == "__main__":
    pytest.main([__file__])
    #test_lof(20, True)
    #test_lof(metric='euclidean')



"""Test DSS functions."""
import os
from tempfile import TemporaryDirectory

import matplotlib.pyplot as plt
import numpy as np
import pytest
from numpy.testing import assert_allclose
from scipy import signal

from meegkit import dss
from meegkit.utils import create_line_data, fold, tscov, unfold

rng = np.random.default_rng(10)


@pytest.mark.parametrize("n_bad_chans", [0, -1])
def test_dss0(n_bad_chans):
    """Test dss0.

    Find the linear combinations of multichannel data that
    maximize repeatability over trials. Data are time * channel * trials.

    Uses dss0().

    `n_bad_chans` set the values of the first corresponding number of channels
    to zero.
    """
    n_samples = 300
    data, source = create_line_data(n_samples=n_samples, n_bad_chans=n_bad_chans)

    # apply DSS to clean them
    c0, _ = tscov(data)
    c1, _ = tscov(np.mean(data, 2))
    [todss, _, pwr0, pwr1] = dss.dss0(c0, c1)
    z = fold(np.dot(unfold(data), todss), epoch_size=n_samples)

    best_comp = np.mean(z[:, 0, :], -1)
    scale = np.ptp(best_comp) / np.ptp(source)

    assert_allclose(np.abs(best_comp), np.abs(np.squeeze(source)) * scale,
                    atol=1e-6)  # use abs as DSS component might be flipped


def test_dss1(show=True):
    """Test DSS1 (evoked)."""
    n_samples = 300
    data, source = create_line_data(n_samples=n_samples, fline=.01)

    todss, _, pwr0, pwr1 = dss.dss1(data, weights=None)
    z = fold(np.dot(unfold(data), todss), epoch_size=n_samples)

    best_comp = np.mean(z[:, 0, :], -1)
    scale = np.ptp(best_comp) / np.ptp(source)

    assert_allclose(np.abs(best_comp), np.abs(np.squeeze(source)) * scale,
                    atol=1e-6)  # use abs as DSS component might be flipped

    # With weights
    weights = np.zeros(n_samples)
    weights[100:200] = 1  # we placed the signal is in the middle of the trial
    todss, _, pwr0, pwr1 = dss.dss1(data, weights=weights)
    z = fold(np.dot(unfold(data), todss), epoch_size=n_samples)

    best_comp = np.mean(z[:, 0, :], -1)
    scale = np.ptp(best_comp) / np.ptp(source)

    if show:
        f, (ax1, ax2, ax3) = plt.subplots(3, 1)
        ax1.plot(source, label="source")
        ax2.plot(np.mean(data, 2), label="data")
        ax3.plot(best_comp, label="recovered")
        plt.legend()
        plt.show()

    assert_allclose(np.abs(best_comp), np.abs(np.squeeze(source)) * scale,
                    atol=1e-6)  # use abs as DSS component might be flipped


@pytest.mark.parametrize("nkeep", [None, 2])
def test_dss_line(nkeep):
    """Test line noise removal."""
    sr = 200
    fline = 20
    nsamples = 10000
    nchans = 10
    s = create_line_data(n_samples=3 * nsamples, n_chans=nchans,
                         n_trials=1, fline=fline / sr, SNR=2)[0][..., 0]

    def _plot(x):
        f, ax = plt.subplots(1, 2, sharey=True)
        f, Pxx = signal.welch(x, sr, nperseg=1024, axis=0,
                              return_onesided=True)
        ax[1].semilogy(f, Pxx)
        f, Pxx = signal.welch(s, sr, nperseg=1024, axis=0,
                              return_onesided=True)
        ax[0].semilogy(f, Pxx)
        ax[0].set_xlabel("frequency [Hz]")
        ax[1].set_xlabel("frequency [Hz]")
        ax[0].set_ylabel("PSD [V**2/Hz]")
        ax[0].set_title("before")
        ax[1].set_title("after")
        plt.show()


    # 2D case, n_outputs == 1
    out, _ = dss.dss_line(s, fline, sr, nkeep=nkeep)
    _plot(out)

    # Test blocksize
    out, _ = dss.dss_line(s, fline, sr, nkeep=nkeep, blocksize=1000)
    _plot(out)

    # Test n_outputs > 1
    out, _ = dss.dss_line(s, fline, sr, nkeep=nkeep, nremove=2)
    # _plot(out)

    # Test n_trials > 1
    x = rng.standard_normal((nsamples, nchans, 4))
    artifact = np.sin(
        np.arange(nsamples) / sr * 2 * np.pi * fline)[:, None, None]
    artifact[artifact < 0] = 0
    artifact = artifact ** 3
    s = x + 10 * artifact
    out, _ = dss.dss_line(s, fline, sr, nremove=1)
    plt.close("all")

def test_dss_line_iter():
    """Test line noise removal."""

    # data = np.load("data/dss_line_iter_test_data.npy")
    # # time x channel x trial sf=200 fline=50

    sr = 200
    fline = 25
    n_samples = 9000
    n_chans = 10

    # 2D case, n_outputs == 1
    x, _ = create_line_data(n_samples, n_chans=n_chans, n_trials=1,
                            noise_dim=10, SNR=2, fline=fline / sr)
    x = x[..., 0]

    # RuntimeError when max iterations has been reached
    with pytest.raises(RuntimeError):
        out, _ = dss.dss_line_iter(x, fline + 1, sr,
                                   show=False, n_iter_max=2)

    with TemporaryDirectory() as tmpdir:
        out, _ = dss.dss_line_iter(x, fline + .5, sr,
                                   prefix=os.path.join(tmpdir, "dss_iter_"),
                                   show=True)

    def _plot(before, after):
        f, ax = plt.subplots(1, 2, sharey=True)
        f, Pxx = signal.welch(before[:, -1], sr, nperseg=1024, axis=0,
                              return_onesided=True)
        ax[0].semilogy(f, Pxx)
        f, Pxx = signal.welch(after[:, -1], sr, nperseg=1024, axis=0,
                              return_onesided=True)
        ax[1].semilogy(f, Pxx)
        ax[0].set_xlabel("frequency [Hz]")
        ax[1].set_xlabel("frequency [Hz]")
        ax[0].set_ylabel("PSD [V**2/Hz]")
        ax[0].set_title("before")
        ax[1].set_title("after")
        plt.show()

    _plot(x, out)

    # # Test n_trials > 1 TODO
    x, _ = create_line_data(n_samples, n_chans=n_chans, n_trials=2,
                            noise_dim=10, SNR=2, fline=fline / sr)
    out, _ = dss.dss_line_iter(x, fline, sr, show=False)
    plt.close("all")


def profile_dss_line(nkeep):
    """Test line noise removal."""
    import cProfile
    import io
    from pstats import SortKey, Stats

    sr = 200
    fline = 20
    nsamples = 1000000
    nchans = 99
    s = create_line_data(n_samples=3 * nsamples, n_chans=nchans,
                         n_trials=1, fline=fline / sr, SNR=2)[0][..., 0]

    pr = cProfile.Profile()
    pr.enable()
    out, _ = dss.dss_line(s, fline, sr, nkeep=nkeep)
    pr.disable()
    s = io.StringIO()
    sortby = SortKey.CUMULATIVE
    ps = Stats(pr, stream=s).sort_stats(sortby)
    ps.print_stats()
    print(s.getvalue())

if __name__ == "__main__":
    pytest.main([__file__])
    # create_data(SNR=5, show=True)
    # test_dss1(True)
    # test_dss_line(2)
    # test_dss_line_iter()
    # profile_dss_line(None)



"""Rhythmic Entrainment Source Separation."""
import warnings

import numpy as np
from scipy import linalg
from sklearn.base import BaseEstimator, TransformerMixin

from meegkit.utils import demean, gaussfilt, mrdivide, tscov


class RESS(TransformerMixin, BaseEstimator):
    """Rhythmic Entrainment Source Separation.

    As described in [1]_.

    Parameters
    ----------
    sfreq : int
        Sampling frequency.
    peak_freq : float
        Peak frequency.
    neig_freq : float
        Distance of neighbouring frequencies away from peak frequency, +/- in
        Hz (default=1).
    peak_width : float
        FWHM of the peak frequency (default=.5).
    neig_width : float
        FWHM of the neighboring frequencies (default=1).
    n_keep : int
        Number of components to keep (default=1). -1 keeps all components.
    gamma : float
        Regularization coefficient, between 0 and 1 (default=0.01, which
        corresponds to 1 % regularization and helps reduce numerical problems
        for noisy or reduced-rank matrices [2]_).
    compute_unmixing : bool
        If True, also computing unmixing matrices (from_ress), used to
        transform the data back into sensor space.

    Examples
    --------
    To project the RESS components back into sensor space, one can proceed as
    follows:
    >>> # First create RESS estimator and fit_transform the data
    >>> r = ress.RESS(sfreq, peak_freq, compute_unmixing=True)
    >>> out = r.fit_transform(data)
    >>> # Then matrix multiply each trial by the unmixing matrix:
    >>> fromRESS = r.from_ress
    >>> proj = matmul3d(out, fromRESS)

    To transform a new observation into RESS component space (e.g. in the
    context of a cross-validation, with separate train/test sets) use the
    `transform` method:
    >>> new_comp = r.transform(newdata)

    References
    ----------
    .. [1] Cohen, M. X., & Gulbinaite, R. (2017). Rhythmic entrainment source
       separation: Optimizing analyses of neural responses to rhythmic sensory
       stimulation. Neuroimage, 147, 43-56.
    .. [2] Cohen, M. X. (2021). A tutorial on generalized eigendecomposition
       for source separation in multichannel electrophysiology.
       ArXiv:2104.12356 [Eess, q-Bio].
    """

    def __init__(self, sfreq: int, peak_freq: float, neig_freq: float = 1,
                 peak_width: float = 0.5, neig_width: float = 1, n_keep: int = 1,
                 gamma: float = 0.01, compute_unmixing: bool = False):

        self.sfreq = sfreq
        self.peak_freq = peak_freq
        self.neig_freq = neig_freq
        self.peak_width = peak_width
        self.neig_width = neig_width
        self.n_keep = n_keep
        self.gamma = gamma
        self.compute_unmixing = compute_unmixing

    def fit(self, X, y=None):
        """Learn a RESS filter.

        X : np.array (n_samples, n_chans, n_trials)
            Follow MNE format
        y : (ignored)
            Ignored parameter.

        Returns
        -------
        self : object
            RESS class instance.
        """
        if X.ndim == 2:
            X = X[np.newaxis, :, :]
            warnings.warn("Fitting the RESS on only one sample !")

        _, n_chans, _ = X.shape

        # Compute mean along epoch + trial and remove it
        X = demean(X)

        if self.n_keep == -1:
            self.n_keep = n_chans

        # Covariance of  neighbor frequencies (noise)
        c01 = tscov(gaussfilt(X, self.sfreq,  self.peak_freq + self.neig_freq,
                              fwhm=self.neig_width, n_harm=1))[0]
        c02 = tscov(gaussfilt(X, self.sfreq, self.peak_freq - self.neig_freq,
                              fwhm=self.neig_width, n_harm=1))[0]

        # Covariance of the signal
        c1 = tscov(gaussfilt(X, self.sfreq, self.peak_freq, fwhm=self.peak_width,
                             n_harm=1))[0]

        # add 1% regularization to avoid numerical precision problems in the GED
        c0 = (c01 + c02) / 2
        c0 = c0 * (1 - self.gamma) + self.gamma * np.trace(c0) / len(c0) * np.eye(len(c0))

        # Perform generalized eigendecomposition. Solves a Generalized
        # Eigenvalue Problem: find a vector `to_ress` that maximize the ratio
        # c0^{-1} c1 (max c1 and min c0)
        d, to_ress = linalg.eigh(c1, c0)

        # Keep only the real part as c1 and c0 are symmetric (then PSD) the
        # imaginary part is only a numerical error
        d = d.real
        to_ress = to_ress.real

        # Sort eigenvectors by decreasing eigenvalues. We are looking for the
        # eigenvectors associated with the larger eigenvalue
        idx = np.argsort(d)[::-1]
        d = d[idx]
        to_ress = to_ress[:, idx]

        # Normalize components (yields mixing matrix)
        to_ress /= np.sqrt(np.sum(to_ress, axis=0) ** 2)
        self.to_ress = to_ress[:, np.arange(self.n_keep)]

        if self.compute_unmixing:
            # Compute unmixing matrix
            A = c1 @ to_ress
            B = to_ress.T @ A
            self.from_ress = mrdivide(A, B).T
            self.from_ress = self.from_ress[:self.n_keep, :]

        return self

    def transform(self, X):
        """Project data using the learned RESS filter."""
        if X.ndim == 2:
            out = X @ self.to_ress
        else:
            out = np.zeros((X.shape[0], self.n_keep, X.shape[2]))
            for t in range(X.shape[2]):
                out[..., t] = X[..., t] @ self.to_ress

        return out

    def fit_transform(self, X, y=None):
        """Compute RESS filter and apply it on data."""
        self.fit(X)

        return self.transform(X)

    def inverse_transform(self, X):
        """Backproject the RESS filtered data to the sensors space."""
        return np.matmul(X, self.from_ress)




"""Signal coherence tools.

Compute 2D, 1D, and 0D bicoherence, polycoherence, bispectrum, and polyspectrum.

Bicoherence is a measure of the degree of phase coupling between different
frequency components in a signal. It's essentially a normalized form of the
bispectrum, which itself is a Fourier transform of the third-order cumulant of
a time series:

- 2D bicoherence is the most common form, where one looks at a
  two-dimensional representation of the phase coupling between different
  frequencies. It's a function of two frequency variables.
- 1D bicoherence would imply a slice or a specific condition in the 2D
  bicoherence, reducing it to a function of a single frequency variable. It
  simplifies the analysis by looking at the relationship between a particular
  frequency and its harmonics or other relationships
- 0D bicoherence would imply a single value representing some average or
  overall measure of phase coupling in the signal. It's a highly condensed
  summary, which might represent the average bicoherence over all frequency
  pairs, for instance.


"""
from collections.abc import Iterable

import matplotlib.pyplot as plt
import numpy as np
from numpy.fft import rfft, rfftfreq
from scipy.fftpack import next_fast_len
from scipy.signal import spectrogram


def cross_coherence(x1, x2, sfreq, norm=2, **kwargs):
    """Compute the bispectral cross-coherence between two signals of same length.

    Code adapted from [2]_.

    Parameters
    ----------
    x1: array-like, shape=([n_channels, ]n_samples)
        First signal.
    x2: array-like, shape=([n_channels, ]n_samples)
        Second signal.
    sfreq: float
        Sampling sfreq.
    norm: int | None
        Norm (default=2). If None, return bispectrum.

    Returns
    -------
    f1: array-like
        Frequency axis.
    B: array-like
        Bicoherence between s1 and s2.

    References
    ----------
    .. [1] http://wiki.fusenet.eu/wiki/Bicoherence
    .. [2] https://stackoverflow.com/a/36725871

    """
    f1, _, S1 = compute_spectrogram(x1, sfreq, **kwargs)
    _, _, S2 = compute_spectrogram(x2, sfreq, **kwargs)

    # compute the bicoherence
    ind = np.arange(f1.size // 2)
    indsum = ind[:, None] + ind[None, :]

    P1 = S1[..., ind, None]
    P2 = S2[..., None, ind]
    P12 = S1[..., indsum]

    B = np.mean(P1 * P2 * np.conj(P12), axis=-3)

    if norm is not None: # Bispectrum -> Bicoherence
        B = norm_spectrum(B, P1, P2, P12, time_axis=-3)

    return f1[ind], B


def polycoherence_0d(X, sfreq, freqs, norm=2, synthetic=None, **kwargs):
    """Polycoherence between freqs and sum of freqs.

    Parameters
    ----------
    X: ndarray, shape=(n_channels, n_samples)
        Input data.
    sfreq: float
        Sampling rate.
    freqs: list[float]
        Fixed frequencies.
    norm: int | None
        Norm (default=2).
    synthetic: tuple(float, float, float)
        Used for synthetic signal for some frequencies (freq, amplitude,
        phase), freq must coincide with the first fixed frequency.
    **kwargs: dict
        Additional parameters passed to scipy.signal.spectrogram. Important
        parameters are nperseg, noverlap, nfft.

    Returns
    -------
    B: ndarray, shape=(n_channels,)
        Polycoherence

    """
    assert isinstance(freqs, Iterable), "freqs must be a list"
    freq, t, S = compute_spectrogram(X, sfreq, **kwargs)

    ind = _freq_ind(freq, freqs)
    indsum = _freq_ind(freq, np.sum(freqs))

    Pi = _product_other_freqs(S, ind, synthetic, t)
    Psum = S[..., indsum]

    B = np.mean(Pi * np.conj(Psum), axis=-1)

    if norm is not None:
        # Bispectrum -> Bicoherence
        B = norm_spectrum(B, Pi, 1., Psum, time_axis=-1)

    return B


def polycoherence_1d(X, sfreq, f2, norm=2, synthetic=None, **kwargs):
    """1D polycoherence as a function of f1 and at least one fixed frequency f2.

    Parameters
    ----------
    X: ndarray
        Input data.
    sfreq: float
        Sampling rate
    f2: list[float]
        Fixed frequencies.
    norm: int | None
        Norm (default=2).
    synthetic: tuple(float, float, float)
        Used for synthetic signal for some frequencies (freq, amplitude,
        phase), freq must coincide with the first fixed frequency.
    **kwargs:
        Additional parameters passed to scipy.signal.spectrogram. Important
        parameters are `nperseg`, `noverlap`, `nfft`.

    Returns
    -------
    freq: ndarray, shape=(n_freqs_f1,)
        Frequencies
    B: ndarray, shape=(n_channels, n_freqs_f1)
        1D polycoherence.

    """
    assert isinstance(f2, Iterable), "f2 must be a list"

    f1, t, S = compute_spectrogram(X, sfreq, **kwargs, axis=-1)
    S = np.require(S, "complex64")
    ind2 = _freq_ind(f1, f2)
    ind1 = np.arange(len(f1) - sum(ind2))
    indsum = ind1 + sum(ind2)

    P1 = S[..., ind1]
    Pother = _product_other_freqs(S, ind2, synthetic, t)[..., None]
    Psum = S[..., indsum]

    B = np.nanmean(P1 * Pother * np.conj(Psum), axis=-2)

    if norm is not None:
        B = norm_spectrum(B, P1, Pother, Psum, time_axis=-2)

    return f1[ind1], B


def polycoherence_1d_sum(X, sfreq, fsum, *ofreqs, norm=2, synthetic=None, **kwargs):
    """1D polycoherence with fixed frequency sum fsum as a function of f1.

    Returns polycoherence for frequencies ranging from 0 up to fsum.

    Parameters
    ----------
    X: ndarray
        Input data.
    sfreq: float
        Sampling rate.
    fsum : float
        Fixed frequency sum.
    ofreqs: list[float]
        Fixed frequencies.
    norm: int or None
        If 2 - return polycoherence, n0 = n1 = n2 = 2 (default)
    synthetic: tuple(float, float, float) | None
        Used for synthetic signal for some frequencies (freq, amplitude,
        phase), freq must coincide with the first fixed frequency.

    Returns
    -------
    freq: ndarray, shape=(n_freqs,)
        Frequencies.
    B: ndarray, shape=(n_channels, n_freqs)
        Polycoherence for f1+f2=fsum.

    """
    freq, t, S = compute_spectrogram(X, sfreq, **kwargs)

    indsum = _freq_ind(freq, fsum)
    ind1 = np.arange(np.searchsorted(freq, fsum - np.sum(ofreqs)))
    ind3 = _freq_ind(freq, ofreqs)
    ind2 = indsum - ind1 - sum(ind3)

    P1 = S[..., ind1]
    P2 = S[..., ind2]
    Pother = _product_other_freqs(S, ind3, synthetic, t)[..., None]
    Psum = S[..., [indsum]]

    B = np.mean(P1 * P2 * Pother * np.conj(Psum), axis=-2)

    if norm is not None:
        B = norm_spectrum(B, P1, P2 * Pother, Psum, time_axis=-2)

    return freq[ind1], B


def polycoherence_2d(X, sfreq, ofreqs=None, norm=2, flim1=None, flim2=None,
                     synthetic=None, **kwargs):
    """2D polycoherence between freqs and their sum as a function of f1 and f2.

    2D bicoherence is the most common form, where one looks at a
    two-dimensional representation of the phase coupling between different
    frequencies. It is a function of two frequency variables.

    2D polycoherence as a function of f1 and f2, ofreqs are additional fixed
    frequencies.

    Parameters
    ----------
    X: ndarray
        Input data.
    sfreq: float
        Sampling rate.
    ofreqs: list[float]
        Fixed frequencies.
    norm: int or None
        If 2 - return polycoherence (default), else return polyspectrum.
    flim1: tuple | None
        Frequency limits for f1. If None, it is set to (0, nyquist / 2)
    flim2: tuple | None
        Frequency limits for f2.

    Returns
    -------
    freq1: ndarray, shape=(n_freqs_f1,)
        Frequencies for f1.
    freq2: ndarray, shape=(n_freqs_f2,)
        Frequencies for f2.
    B: ndarray, shape=([n_chans, ]n_freqs_f1, n_freqs_f2)
        Polycoherence.

    """
    freq, t, S = compute_spectrogram(X, sfreq, **kwargs)

    if ofreqs is None:
        ofreqs = []
    if flim1 is None:
        flim1 = (0, (np.max(freq) - np.sum(ofreqs)) / 2)
    if flim2 is None:
        flim2 = (0, (np.max(freq) - np.sum(ofreqs)) / 2)

    # indices ranges matching flim1 and flim2
    ind1 = np.arange(*np.searchsorted(freq, flim1))
    ind2 = np.arange(*np.searchsorted(freq, flim2))
    ind3 = _freq_ind(freq, ofreqs)
    indsum = ind1[:, None] + ind2[None, :] + sum(ind3)

    Pother = _product_other_freqs(S, ind3, synthetic, t)[..., None, None]

    P1 = S[..., ind1, None]
    P2 = S[..., None, ind2] * Pother
    P12 = S[..., indsum]

    # Average over time to get the bispectrum
    B = np.nanmean(P1 * P2 * np.conj(P12), axis=-3)

    if norm is not None: # Bispectrum -> Bicoherence
        B = norm_spectrum(B, P1, P2, P12, time_axis=-3)

    return freq[ind1], freq[ind2], B


def compute_spectrogram(X, sfreq, **kwargs):
    """Compute the complex spectrogram of X.

    Simple wrapper around scipy.signal.spectrogram.

    Returns
    -------
    f: ndarray, shape=(n_freqs,)
        Positive frequencies.
    t: ndarray, shape=(n_timesteps,)
        Time axis.
    S: ndarray, shape=(n_chans, n_timesteps, n_freqs)
        Complex spectrogram (one-sided).
    """
    N = X.shape[-1]
    kwargs.setdefault("nperseg", N // 2)
    kwargs.setdefault("noverlap", N // 5)
    kwargs.setdefault("nfft", next_fast_len(N))

    f, t, S = spectrogram(X, fs=sfreq, mode="complex", return_onesided=False, **kwargs)
    S = S[:len(f) // 2]  # only positive frequencies
    f = f[:len(f) // 2]  # only positive frequencies
    S = np.require(S, "complex64")
    S = np.swapaxes(S, -1, -2)  # transpose (f, t) -> (t, f)
    return f, t, S


def norm_spectrum(spec, P1, P2, P12, time_axis=-2):
    """Compute bicoherence from bispectrum.

    For formula see [1]_.

    Parameters
    ----------
    spec: ndarray, shape=(n_chans, n_freqs)
        Polyspectrum.
    P1: array-like, shape=(n_chans, n_times, n_freqs_f1)
        Spectrum evaluated at f1.
    P2: array-like, shape=(n_chans, n_times, n_freqs_f2)
        Spectrum evaluated at f2.
    P12: array-like, shape=(n_chans, n_times, n_freqs)
        Spectrum evaluated at f1 + f2.
    time_axis: int
        Time axis.

    Returns
    -------
    coh: ndarray, shape=(n_chans,)
        Polycoherence.

    References
    ----------
    .. [1] https://en.wikipedia.org/wiki/Bicoherence

    """
    coh = np.abs(spec) ** 2
    norm = np.nanmean(np.abs(P1 * P2) ** 2, axis=time_axis)
    norm *= np.nanmean(np.abs(np.conj(P12)) ** 2, axis=time_axis)
    coh /= norm
    coh **= 0.5
    return coh


def plot_polycoherence(freq1, freq2, bicoh, ax=None):
    """Plot polycoherence."""
    df1 = freq1[1] - freq1[0]  # resolution
    df2 = freq2[1] - freq2[0]
    freq1 = np.append(freq1, freq1[-1] + df1) - 0.5 * df1
    freq2 = np.append(freq2, freq2[-1] + df2) - 0.5 * df2

    if ax is None:
        f, ax = plt.subplots()

    ax.pcolormesh(freq2, freq1, np.abs(bicoh))
    ax.set_xlabel("freq (Hz)")
    ax.set_ylabel("freq (Hz)")
    # ax.colorbar()
    return ax

def plot_polycoherence_1d(freq, coh):
    """Plot polycoherence for fixed frequencies."""
    plt.figure()
    plt.plot(freq, coh)
    plt.xlabel("freq (Hz)")


def plot_signal(t, signal, ax=None):
    """Plot signal and spectrum."""
    if ax is None:
        f, ax = plt.subplots(2, 1)

    ax[0].plot(t, signal)
    ax[0].set_xlabel("time (s)")

    ndata = len(signal)
    nfft = next_fast_len(ndata)
    freq = rfftfreq(nfft, t[1] - t[0])
    spec = rfft(signal, nfft) * 2 / ndata
    ax[1].plot(freq, np.abs(spec))
    ax[1].set_xlabel("freq (Hz)")
    return ax

def _freq_ind(freq, f0):
    """Find the index of the frequency closest to f0."""
    if isinstance(f0, Iterable):
        return [np.argmin(np.abs(freq - f)) for f in f0]
    else:
        return np.argmin(np.abs(freq - f0))


def _product_other_freqs(spec, indices, synthetic=None, t=None):
    """Product of all frequencies."""
    if synthetic is None:
        synthetic = ()

    p1 = synthetic_signal(t, synthetic)
    p2 = np.prod(spec[..., indices[len(synthetic):]], axis=-1)
    return p1 * p2


def synthetic_signal(t, synthetic):
    """Create a synthetic complex signal spectrum.

    Parameters
    ----------
    t: array-like
        Time.
    synthetic: list[tuple(float, float, float)]
        List of tuples with (freq, amplitude, phase).

    Returns
    -------
    Complex signal.

    """
    return np.prod([amp * np.exp(2j * np.pi * freq * t + phase)
                    for (freq, amp, phase) in synthetic], axis=0)


import matplotlib.pyplot as plt
import pytest


def pytest_addoption(parser):
    """Add command line option to pytest."""
    parser.addoption(
        "--runslow",
        action="store_true",
        default=False,
        help="run slow tests"
    )
    parser.addoption(
        "--noplots",
        action="store_true",
        default=False,
        help="halt on plots"
    )


def pytest_collection_modifyitems(config, items):
    """Do not skip slow test if option provided."""
    if config.getoption("--noplots"):
        plt.switch_backend("agg")

    if config.getoption("--runslow"):
        # --runslow given in cli: do not skip slow tests
        return
    skip_slow = pytest.mark.skip(reason="need --runslow option to run")
    for item in items:
        if "slow" in item.keywords:
            item.add_marker(skip_slow)



"""Audio and signal processing tools."""
import numpy as np
import scipy.signal as ss
from scipy.linalg import lstsq, solve, toeplitz
from scipy.signal import lfilter

from .covariances import convmtx


def modulation_index(phase, amp, n_bins=18):
    """Compute the Modulation Index (MI) between two signals.

    MI is a measure of the amount of phase-amplitude coupling. Phase angles are
    expected to be in radians [1]_. MI is derived from the Kullbach-Leibner
    distance, a measure for the disparity of two distributions, which is also
    returned here. MI is recommend the modulation index for noisy and short
    data epochs with unknown forms of coupling [2]_.

    Parameters
    ----------
    phase : array
        Phase vector, in radians.
    amp : array
        Amplitude vector.
    n_bins : int
        Number of bins in which to discretize phase (default=18 bins, giving
        a 20-degree resolution).

    Returns
    -------
    MI : array
        Tort's Modulation index.
    KL : array
        Kullbach-Leibner distance.

    Examples
    --------
    >> phas = np.random.rand(100, 1) * 2 * np.pi - np.pi
    >> ampl = rng.standard_normal((100, 1) * 30 + 100
    >> MI, KL = modulation_index(phas, ampl)

    Notes
    -----
    Phase and amplitude can be derived directly from any time series through
    the analytic signal:
    >> analytic_signal = hilbert(filtered_data)
    >> phase = np.phase(analytic_signal)
    >> amplitude = np.abs(analytic_signal)

    MI can be subjected to permutation testing to assess significance. For
    permutation testing, the observed coupling value is compared to a
    distribution of shuffled coupling values. Shuffled coupling values are
    constructed by calculating the coupling value between the original phase
    time series and a permuted amplitude time series. The permuted amplitude
    time series can be constructed by cutting the amplitude time series at a
    random time point and reversing the order of both parts [2]_. The observed
    coupling value is standardized to the distribution of the shuffled coupling
    values according to the following formula:

    MI_z = (MI_observed − µ_MI_shuffled) / σ_MI_shuffled

    where μ denotes the mean and σ the standard deviation. Only when the
    observed phase-locking value is larger than 95% of shuffled values, it is
    defined as significant. See [2]_ for details.

    References
    ----------
    .. [1] Tort AB, Komorowski R, Eichenbaum H, Kopell N. Measuring
       phase-amplitude coupling between neuronal oscillations of different
       frequencies. J Neurophysiol. 2010 Aug;104(2):1195-210. doi:
       10.1152/jn.00106.2010.
    .. [2] Huelsemann, M. J., Naumann, E., & Rasch, B. (2018). Quantification
       of Phase-Amplitude Coupling in Neuronal Oscillations: Comparison of
       Phase-Locking Value, Mean Vector Length, and Modulation Index. bioRxiv,
       290361.

    """
    phase = phase.squeeze()
    amp = amp.squeeze()
    if phase.shape != amp.shape or phase.ndims > 1 or amp.ndims:
        raise AttributeError("Inputs must be 1D vectors of same length.")

    # Convert phase to degrees
    phasedeg = np.degrees(phase)

    # Calculate mean amplitude in each phase bin
    binsize = 360 / n_bins
    phase_lo = np.arange(-180, 180, binsize)
    mean_amp = np.zeros(len(phase_lo))
    for b in range(len(phase_lo)):
        phaserange = np.logical_and(phasedeg >= phase_lo[b],
                                    phasedeg < (phase_lo[b] + binsize))
        mean_amp[b] = np.mean(amp[phaserange])

    # Compute the probability of an amplitude unit being in a phase bin
    p_j = mean_amp / np.sum(mean_amp)

    # Get a meaningful KL distance when observed probability in a bin is 0
    if np.any(p_j == 0):
        p_j[p_j == 0] = np.finfo(float).eps

    # Phase-amplitude coupling is defined by a distribution that significantly
    # deviates from the uniform distribution. Kullback-Leibler distance is
    # calculated by the following formula: KL = log(N) − H(p), where H is
    # Shannon entropy, and N is the number of bins.
    H = -np.sum(p_j * np.log10(p_j))
    Hmax = np.log10(n_bins)
    KL = Hmax - H
    MI = KL / Hmax

    return MI, KL


def smooth(x, window_len, window="square", axis=0, align="left"):
    """Smooth a signal using a window with requested size along a given axis.

    This method is based on the convolution of a scaled window with the signal.
    The signal is prepared by introducing reflected copies of the signal (with
    the window size) in both ends so that transient parts are minimized in the
    beginning and end part of the output signal.

    Parameters
    ----------
    x : array
        The input signal.
    window_len : float
        The dimension of the smoothing window (in samples). Can be fractionary.
    window : str
        The type of window from 'flat', 'hanning', 'hamming', 'bartlett',
        'blackman' flat window will produce a moving average smoothing.
    axis : int
        Axis along which smoothing will be applied (default=0).
    align : {'left' | 'center'}
        If `left` (default), the convolution if computed in a causal way by
        shifting the output of a normal convolution by the kernel size. If
        `center`, the center of the impulse is used as the location where the
        convolution is summed.

    Returns
    -------
    y : array
        The smoothed signal.

    Examples
    --------
    >> t = linspace(-2, 2, 0.1)
    >> x = sin(t) + randn(len(t)) * 0.1
    >> y = smooth(x, 2)

    See Also
    --------
    np.hanning, np.hamming, np.bartlett, np.blackman, np.convolve,
    scipy.signal.lfilter

    Notes
    -----
    length(output) != length(input), to correct this, we return :
    >> y[(window_len / 2 - 1):-(window_len / 2)]  # noqa
    instead of just y.

    """
    if x.shape[axis] < window_len:
        raise ValueError("Input vector needs to be bigger than window size.")
    if window not in ["square", "hanning", "hamming", "bartlett", "blackman"]:
        raise ValueError("Unknown window type.")
    if window_len == 0:
        raise ValueError("Smoothing kernel must be at least 1 sample wide")
    if window_len == 1:
        return x

    def _smooth1d(x, n, align="left"):
        if x.ndim != 1:
            raise ValueError("Smooth only accepts 1D arrays")

        frac, n = np.modf(n)
        n = int(n)

        if window == "square":  # moving average
            w = np.ones(n, "d")
            w = np.r_[w, frac]
        else:
            w = eval("np." + window + "(n)")

        if align == "center":
            a = x[n - 1:0:-1]
            b = x[-2:-n - 1:-1]
            s = np.r_[a, x, b]
            out = np.convolve(w / w.sum(), s, mode="same")
            return out[len(a):-len(b)]

        elif align == "left":
            out = ss.lfilter(w / w.sum(), 1, x)
            return out

    if x.ndim > 1:  # apply along given axis
        y = np.apply_along_axis(_smooth1d, axis, x, n=window_len)
    else:
        y = _smooth1d(x, n=np.abs(window_len))

    return y


def lowpass_env_filtering(x, cutoff=150., n=1, sfreq=22050):
    """Low-pass filters a signal using a Butterworth filter.

    Parameters
    ----------
    x : ndarray
    cutoff : float, optional
        Cut-off frequency of the low-pass filter, in Hz. The default is 150 Hz.
    n : int, optional
        Order of the low-pass filter. The default is 1.
    sfreq : float, optional
        Sampling frequency of the signal to filter. The default is 22050 Hz.

    Returns
    -------
    ndarray
        Low-pass filtered signal.

    """
    b, a = ss.butter(N=n, Wn=cutoff * 2. / sfreq, btype="lowpass")
    return ss.lfilter(b, a, x)


def hilbert_envelope(x):
    """Calculate the Hilbert envelope of a signal.

    Parameters
    ----------
    x : array
        Signal on which to calculate the hilbert envelope. The calculation
        is done along the last axis (i.e. ``axis=-1``).

    Returns
    -------
    ndarray

    """
    def next_pow_2(x):
        return 1 if x == 0 else 2**(x - 1).bit_length()

    signal = np.asarray(x)
    N_orig = signal.shape[-1]
    # Next power of 2.
    N = next_pow_2(N_orig)
    y_h = ss.hilbert(signal, N)
    # Return signal with same shape as original
    return np.abs(y_h[..., :N_orig])


def spectral_envelope(x, sfreq, lowpass=32):
    """Compute envelope with convolution.

    Notes
    -----
    The signal is first padded to avoid edge effects. To align the envelope
    with the input signal, we return :
    >> y[(window_len / 2 - 1):-(window_len / 2)]  # noqa

    """
    x = np.squeeze(x)
    if x.ndim > 1:
        raise AttributeError("x must be 1D")
    if lowpass is None:
        lowpass = sfreq / 2

    # Pad signal with reflection
    win = sfreq // lowpass  # window size in samples
    a = x[win - 1:0:-1]
    b = x[-2:-win - 1:-1]
    s = np.r_[a, x, b]

    # Convolve squared signal with a square window and take cubic root
    y = np.convolve(s ** 2, np.ones((win,)) / win, mode="same") ** (1 / 3)
    return y[len(a):-len(b)]


def gaussfilt(data, sfreq, f, fwhm, n_harm=1, shift=0, return_empvals=False, show=False):
    """Narrow-band filter via frequency-domain Gaussian.

    Empirical frequency and FWHM depend on the sampling rate and the number of
    time points, and may thus be slightly different from the requested values.

    Parameters
    ----------
    data : ndarray
        EEG data, shape=(n_samples, n_channels[, ...])
    sfreq : int
        Sampling rate in Hz.
    f : float
        Break frequency of filter.
    fhwm : float
        Standard deviation of filter, defined as full-width at half-maximum
        in Hz.
    n_harm : int
        Number of harmonics of the frequency to consider.
    shift : int
        Amount shift peak frequency by (only useful when considering harmonics,
        otherwise leave to 0).
    return_empvals : bool
        Return empirical values (default: False).
    show : bool
        Set to True to show the frequency-domain filter shape.

    Returns
    -------
    filtdat : ndarray
        Filtered data.
    empVals : float
        The empirical frequency and FWHM.
    """
    # input check
    assert (data.shape[1] <= data.shape[0]
            ), "n_channels must be less than n_samples"
    assert ((f - fwhm) >= 0), "increase frequency or decrease FWHM"
    assert (fwhm >= 0), "FWHM must be greater than 0"

    # frequencies
    hz = np.fft.fftfreq(data.shape[0], 1. / sfreq)
    empVals = np.zeros((2,))

    # compute empirical frequency and standard deviation
    idx_p = np.searchsorted(hz[hz >= 0], f, "left")

    # create Gaussian
    fx = np.zeros_like(hz)
    for i_harm in range(1, n_harm + 1):  # make one gaussian per harmonic
        s = fwhm * (2 * np.pi - 1) / (4 * np.pi)  # normalized width
        x = hz.copy()
        x -= (f * i_harm - shift)
        gauss = np.exp(-.5 * (x / s)**2)  # gaussian
        gauss = gauss / np.max(gauss)  # gain-normalized
        fx = fx + gauss

    # create Gaussian
    for i_harm in range(1, n_harm + 1):  # make one gaussian per harmonic
        s = fwhm * (2 * np.pi - 1) / (4 * np.pi)  # normalized width
        x = hz.copy()
        x += (f * i_harm - shift)
        gauss = np.exp(-.5 * (x / s) ** 2)  # gaussian
        gauss = gauss / np.max(gauss)  # gain-normalized
        fx = fx + gauss

    # filter
    if data.ndim == 2:
        filtdat = 2 * np.real(np.fft.ifft(
                              np.fft.fft(data, axis=0) * fx[:, None], axis=0))
    elif data.ndim == 3:
        filtdat = 2 * np.real(np.fft.ifft(
                              np.fft.fft(data, axis=0) * fx[:, None, None],
                              axis=0))

    if return_empvals or show:
        empVals[0] = hz[idx_p]
        # find values closest to .5 after MINUS before the peak
        empVals[1] = hz[idx_p - 1 + np.searchsorted(fx[:idx_p], 0.5)] \
            - hz[np.searchsorted(fx[:idx_p + 1], 0.5)]

    if show:
        # inspect the Gaussian (turned off by default)
        import matplotlib.pyplot as plt
        plt.figure(1)
        plt.plot(hz, fx, "o-")
        plt.xlim([0, None])

        title = f"Requested: {f}, {fwhm} Hz\nEmpirical: {empVals[0]}, {empVals[1]} Hz"
        plt.title(title)
        plt.xlabel("Frequency (Hz)")
        plt.ylabel("Amplitude gain")
        plt.show()

    if return_empvals:
        return filtdat, empVals
    else:
        return filtdat


def teager_kaiser(x, m=1, M=1, axis=0):
    """Mean Teager-Kaiser energy operator.

    The discrete version of the Teager-Kaiser operator is computed according
    to:

    y[n] = x[n] ** {2 / m} - (x[n - M] * x[n + M]) ** {1 / m}

    with m the exponent parameter and M the lag parameter which both are
    usually equal to 1 for a conventional operator. The Teaser-Kaiser operator
    can be used to track amplitude modulations (AM) and/or frequency
    modulations (FM).

    Parameters
    ----------
    x : array, shape=(n_samples[, n_channels][, n_trials])
        Input data.
    m : int
        Exponent parameter.
    M : int
        Lag parameter.
    axis : int
        Axis to compute metric on.

    Returns
    -------
    array, shape=(n_samples - 2 * M[, n_channels][, n_trials])
        Instantaneous energy.

    References
    ----------
    Adapted from the TKEO function in R library `seewave`.

    Examples
    --------
    >>> x = np.array([1,  3, 12, 25, 10])
    >>> tk_energy = teager_kaiser(x)

    """
    def tg(x, M, m):
        return x[M:-M] ** (2 / m) - (x[2 * M:] * x[:-2 * M]) ** (1 / m)

    return np.apply_along_axis(tg, axis, x, M, m)


def slope_sum(x, w: int, axis=0):
    r"""Slope sum function.

    The discrete version of the Teager-Kaiser operator is computed according
    to:

    y[n] = \\sum_{i=k-w}^k (y_i -y_{i-1})

    Parameters
    ----------
    x : array, shape=(n_samples[, n_channels][, n_trials])
        Input data.
    w : int
        Window.

    References
    ----------
    https://ieeexplore.ieee.org/document/8527395

    """
    def ss(x, w):
        out = np.diff(x, prepend=0)
        out = out.cumsum()
        out[w:] = out[w:] - out[:-w]
        return out

    return np.apply_along_axis(ss, axis, x, w)


def stmcb(x, u_in=None, q=None, p=None, niter=5, a_in=None):
    """Compute linear model via Steiglitz-McBride iteration.

    [B,A] = stmcb(H,NB,NA) finds the coefficients of the system B(z)/A(z) with
    approximate impulse response H, NA poles and NB zeros.

    [B,A] = stmcb(H,NB,NA,N) uses N iterations.  N defaults to 5.

    [B,A] = stmcb(H,NB,NA,N,Ai) uses the vector Ai as the initial
    guess at the denominator coefficients. If you don't specify Ai,
    STMCB uses [B,Ai] = PRONY(H,0,NA) as the initial conditions.

    [B,A] = STMCB(X,Y,NB,NA,N,Ai) finds the system coefficients B and
    A of the system which, given X as input, has Y as output. N and Ai
    are again optional with default values of N = 5, [B,Ai] = PRONY(Y,0,NA).
    Y and X must be the same length.

    Parameters
    ----------
    x : array
    u_in : array
    q : int
    p : int
    n_iter : int
    a_in : array

    Returns
    -------
    b : array
        Filter coefficients (denominator).
    a : array
        Filter coefficients (numerator).

    Examples
    --------
    Approximate the impulse response of a Butterworth filter with a
    system of lower order:

    >>> [b, a] = butter(6, 0.2)                # Butterworth filter design
    >>> h = filter(b, a, [1, zeros(1,100)])    # Filter data using above filter
    >>> freqz(b, a, 128)                       # Frequency response
    >>> [bb, aa] = stmcb(h, 4, 4)
    >>> plt.plot(freqz(bb, aa, 128))

    References
    ----------
    Authors: Jim McClellan, 2-89 T. Krauss, 4-22-93, new help and options
    Copyright 1988-2004 The MathWorks, Inc.

    """
    if u_in is None:
        if q is None:
            q = 0
        if a_in is None:
            a_in, _ = prony(x, 0, p)

        # make a unit impulse whose length is same as x
        u_in = np.zeros(len(x))
        u_in[0] = 1.
    else:
        if len(u_in) != len(x):
            raise ValueError(
                f"stmcb: u_in and x must be of the same size: {len(u_in)} != {len(x)}")
        if a_in is None:
            q = 0
            _, a_in = prony(x, q, p)

    a = a_in
    N = len(x)
    for _i in range(niter):
        u = lfilter([1], a, x)
        v = lfilter([1], a, u_in)
        C1 = convmtx(u, (p + 1)).T
        C2 = convmtx(v, (q + 1)).T
        T = np.hstack((-C1[0:N, :], C2[0:N, :]))

        # move 1st column to RHS and do least-squares
        # c = T(:,2:p+q+2)\( -T(:,1));
        #
        # If not squared matrix: numpy.linalg.lstsq
        # If squared matrix: numpy.linalg.solve
        T_left = T[:, 1:p + q + 2]
        T_right = -T[:, 0]
        if T.shape[0] != T.shape[1]:
            c, residuals, rank, singular_values = lstsq(
                T_left, T_right)  # lstsq in python returns more stuff
        else:
            c = solve(T_left, T_right)

        # denominator coefficients
        a_left = np.array([1])
        a_right = c[:p]
        a = np.hstack((a_left, a_right))

        # numerator coefficients
        b = c[p:p + q + 1]

    a = a.T
    b = b.T
    return b, a


def prony(h, nb, na):
    """Prony's method for time-domain IIR filter design.

    [B,A] = PRONY(H, NB, NA) finds a filter with numerator order NB,
    denominator order NA, and having the impulse response in vector H. The IIR
    filter coefficients are returned in length NB+1 and NA+1 row vectors B and
    A, ordered in descending powers of Z.  H may be real or complex.

    If the largest order specified is greater than the length of H, H is padded
    with zeros.

    Parameters
    ----------
    h : array
        Impulse response.
    nb : int
        Numerator order.
    na : int
        Denominator order.

    References
    ----------
    T.W. Parks and C.S. Burrus, Digital Filter Design, John Wiley and Sons,
    1987, p226.

    Notes
    -----
    Copyright 1988-2012 The MathWorks, Inc.

    """
    K = len(h) - 1
    M = nb
    N = na
    if K <= max(M, N):
        # zero-pad input if necessary
        K = max(M, N) + 1
        h[K + 1] = 0

    c = h[0]
    if c == 0:
        c = 1  # avoid division by zero

    H = toeplitz(h / c, np.array(np.hstack((1, np.zeros(K)))))

    # K+1 by N+1
    if K > N:
        # Here we are just getting rid of all the columns after N+1
        H = H[:, :N + 1]

    # Partition H matrix
    H1 = H[:M + 1, :]

    # M+1 by N+1
    h1 = H[M + 1:K + 1, 0]

    # K-M by 1
    H2 = H[M + 1:K + 1, 1:N + 1]

    # K-M by N
    a_right = np.linalg.lstsq(-H2, h1, rcond=None)[0]
    a = np.r_[(np.array([1]), a_right)][None, :]
    b = np.dot(np.dot(c, a), H1.T)
    return b, a



import matplotlib.pyplot as plt
import numpy as np
from scipy.signal import hilbert

from meegkit.phase import (
    NonResOscillator,
    ResOscillator,
    locking_based_phase,
)


def test_noisy_signal(show=True):
    """Replicate the figure from the paper."""
    # Model data, all three algorithms
    npt = 100000
    fs = 100
    s0  = generate_multi_comp_data(npt, fs)  # Generate test data
    s = generate_noisy_signal(npt, fs, noise=0.8)
    dt = 1 / fs
    time = np.arange(npt) * dt

    # The test signal s and its Hilbert amplitude ah (red); one can see that
    # ah does not represent a good envelope for s. On the contrary, the
    # Hilbert-based phase estimation yields good results, and therefore we take
    # it for the ground truth.
    gta = np.abs(hilbert(s0))  # ground truth amplitude
    gtp = np.angle(hilbert(s0))  # ground truth phase

    hta = np.abs(hilbert(s))  # Hilbert amplitude
    htp = np.angle(hilbert(s))  # Hilbert phase

    osc = NonResOscillator(fs, 1.1)
    nrp, nra = osc.transform(s)

    osc = ResOscillator(fs, 1.1)
    rp, ra = osc.transform(s)

    if show:
        f, ax = plt.subplots(3, 2, sharex="col", sharey=True, figsize=(12, 8))
        ax[0, 0].plot(time, gtp, lw=.75, label="Ground truth")
        ax[0, 0].plot(time, htp, lw=.75, label=r"$\phi_H$")
        ax[0, 0].set_ylabel(r"$\phi_H$")
        ax[0, 0].set_title("Signal and its Hilbert phase")

        ax[1, 0].plot(time, gtp, lw=.75, label="Ground truth")
        ax[1, 0].plot(time, nrp, lw=.75, label=r"$\\phi_N$")
        ax[1, 0].set_ylabel(r"$\phi_N$")
        ax[1, 0].set_ylim([-np.pi, np.pi])
        ax[1, 0].set_title("Nonresonant oscillator")

        ax[2, 0].plot(time, gtp, lw=.75, label="Ground truth")
        ax[2, 0].plot(time, rp, lw=.75, label=r"$\\phi_N$")
        ax[2, 0].set_ylim([-np.pi, np.pi])
        ax[2, 0].set_ylabel(r"$\phi_H - \phi_R$")
        ax[2, 0].set_xlabel("Time")
        ax[2, 0].set_title("Resonant oscillator")

        ax[0, 1].plot(time, gta, lw=.75, label="Ground truth")
        ax[0, 1].plot(time, hta, lw=.75, label=r"$a_H$")
        ax[0, 1].plot(time, s, lw=.75, label="Signal", color="grey", alpha=.5, zorder=0)
        ax[0, 1].set_ylabel(r"$a_H$")
        ax[0, 1].set_title("Signal and its Hilbert amplitude")

        ax[1, 1].plot(time, gta, lw=.75, label="Ground truth")
        ax[1, 1].plot(time, nra, lw=.75, label=r"$a_N$")
        ax[1, 1].set_ylabel(r"$a_N$")
        ax[1, 1].set_title("Amplitudes")
        ax[1, 1].set_title("Nonresonant oscillator")

        # The resonant oscillator should be much more robust to noise
        ax[2, 1].plot(time, gta, lw=.75, label="Ground truth")
        ax[2, 1].plot(time, ra, lw=.75, label=r"$a_R$")
        ax[2, 1].set_xlabel("Time")
        ax[2, 1].set_ylabel(r"$a_R$")
        ax[2, 1].set_title("Resonant oscillator")
        plt.suptitle("Amplitude (right) and phase (left) - noisy signal")

        ax[2, 0].set_xlim([0, 40])
        ax[2, 1].set_xlim([0, 1000])
        plt.tight_layout()
        plt.show()


def test_all_alg(show=False):
    """Replicate the figure from the paper."""
    # Model data, all three algorithms
    npt = 100000
    fs = 100
    s  = generate_multi_comp_data(npt, fs)  # Generate test data
    dt = 1 / fs
    time = np.arange(npt) * dt

    # The test signal s and its Hilbert amplitude ah (red); one can see that
    # ah does not represent a good envelope for s. On the contrary, the
    # Hilbert-based phase estimation yields good results, and therefore we take
    # it for the ground truth.
    ht_ampl = np.abs(hilbert(s))  # Hilbert amplitude
    ht_phase = np.angle(hilbert(s))  # Hilbert phase

    lb_phase = locking_based_phase(s, dt, npt)
    lb_phi_dif = phase_difference(ht_phase, lb_phase)

    osc = NonResOscillator(fs, 1.1)
    nr_phase, nr_ampl = osc.transform(s)
    nr_phase = nr_phase[:, 0]
    nr_phi_dif = phase_difference(ht_phase, nr_phase)

    osc = ResOscillator(fs, 1.1)
    r_phase, r_ampl = osc.transform(s)
    r_phase = r_phase[:, 0]
    r_phi_dif = phase_difference(ht_phase, r_phase)

    if show:
        f, ax = plt.subplots(4, 2, sharex=True, sharey=True, figsize=(12, 8))
        ax[0, 0].plot(time, s, time, ht_phase, lw=.75)
        ax[0, 0].set_ylabel(r"$s,\phi_H$")
        # ax[0, 0].set_ylim([-2, 3])
        ax[0, 0].set_title("Signal and its Hilbert phase")

        ax[1, 0].plot(time, lb_phi_dif, lw=.75)
        ax[1, 0].axhline(0, color="k", ls=":", zorder=-1)
        ax[1, 0].set_ylabel(r"$\phi_H - \phi_L$")
        ax[1, 0].set_ylim([-np.pi, np.pi])
        ax[1, 0].set_title("Phase locking approach")

        ax[2, 0].plot(time, nr_phi_dif, lw=.75)
        ax[2, 0].axhline(0, color="k", ls=":", zorder=-1)
        ax[2, 0].set_ylabel(r"$\phi_H - \phi_N$")
        ax[2, 0].set_ylim([-np.pi, np.pi])
        ax[2, 0].set_title("Nonresonant oscillator")

        ax[3, 0].plot(time, r_phi_dif, lw=.75)
        ax[3, 0].axhline(0, color="k", ls=":", zorder=-1)
        ax[3, 0].set_ylim([-np.pi, np.pi])
        ax[3, 0].set_ylabel(r"$\phi_H - \phi_R$")
        ax[3, 0].set_xlabel("Time")
        ax[3, 0].set_title("Resonant oscillator")

        ax[0, 1].plot(time, s, time, ht_ampl, lw=.75)
        ax[0, 1].set_ylabel(r"$s,a_H$")
        ax[0, 1].set_title("Signal and its Hilbert amplitude")

        ax[1, 1].axis("off")

        ax[2, 1].plot(time, s, time, nr_ampl, lw=.75)
        ax[2, 1].set_ylabel(r"$s,a_N$")
        ax[2, 1].set_title("Amplitudes")
        ax[2, 1].set_title("Nonresonant oscillator")

        ax[3, 1].plot(time, s, time, r_ampl, lw=.75)
        ax[3, 1].set_xlabel("Time")
        ax[3, 1].set_ylabel(r"$s,a_R$")
        ax[3, 1].set_title("Resonant oscillator")
        plt.suptitle("Amplitude (right) and phase (left) estimation algorithms")
        plt.tight_layout()
        plt.show()

    # Assert that the phase difference between the Hilbert phase and the
    # phase estimated by the locking-based technique is small
    assert np.mean(np.abs(lb_phi_dif)) < 0.27

    # Assert that the phase difference between the Hilbert phase and the
    # phase estimated by the non-resonant oscillator is small
    assert np.mean(np.abs(nr_phi_dif)) < 0.2

    # Assert that the phase difference between the Hilbert phase and the
    # phase estimated by the resonant oscillator is small
    assert np.mean(np.abs(r_phi_dif)) < 0.21


def phase_difference(phi1, phi2):
    cos_phi_dif = np.cos(phi1) * np.cos(phi2) + np.sin(phi1) * np.sin(phi2)
    sin_phi_dif = np.sin(phi1) * np.cos(phi2) - np.cos(phi1) * np.sin(phi2)
    phi_dif = np.arctan2(sin_phi_dif, cos_phi_dif)

    return phi_dif

def generate_multi_comp_data(npt=40000, fs=100):
    """Generate multi-component data with frequency modulation.

    Returns
    -------
    s : np.ndarray
        Generated data.

    """
    dt = 1 / fs
    t = np.arange(1, npt + 1) * dt
    omega1 = np.sqrt(2) / 30  # Frequency of the first component (slow)
    omega2 = np.sqrt(5) / 60  # Frequency of the second component (fast)
    amp = 1 + 0.95 * np.cos(omega1 * t)
    p = t + 5 * np.sin(omega2 * t)
    s = np.cos(p) + 0.2 * np.cos(2 * p + np.pi / 6) + 0.1 * np.cos(3 * p + np.pi / 3)
    s *= amp

    return s


def generate_noisy_signal(npt=40000, fs=100, noise=0.1):
    """Generate multi-component data with frequency modulation.

    Returns
    -------
    s : np.ndarray
        Generated data.

    """
    rng = np.random.default_rng(1)
    # dt = 1 / fs
    # t = np.arange(1, npt + 1) * dt
    s = generate_multi_comp_data(npt, fs)
    s += rng.random(npt) * noise

    return s


if __name__ == "__main__":
    # Run the model_data_all_alg function
    test_all_alg(True)
    # test_noisy_signal(True)


"""Denoising source separation."""
# Authors:  Nicolas Barascud <nicolas.barascud@gmail.com>
#           Maciej Szul <maciej.szul@isc.cnrs.fr>
import numpy as np
from numpy.lib.stride_tricks import sliding_window_view
from scipy import linalg
from scipy.signal import welch

from .tspca import tsr
from .utils import (
    demean,
    gaussfilt,
    matmul3d,
    mean_over_trials,
    pca,
    smooth,
    theshapeof,
    tscov,
    wpwr,
)


def dss1(X, weights=None, keep1=None, keep2=1e-12):
    """DSS to maximise repeatability across trials.

    Evoked-biased DSS denoising.

    Parameters
    ----------
    X: array, shape=(n_samples, n_chans, n_trials)
        Data to denoise.
    weights: array
        Weights.
    keep1: int
        Number of PCs to retain in function:`dss0` (default=all).
    keep2: float
        Ignore PCs smaller than keep2 in function:`dss0` (default=1e-12).

    Returns
    -------
    todss: array, shape=(n_dss_components, n_chans)
        Denoising matrix to convert X to normalized DSS components.
    from: array, shape=(n_dss_components, n_chans)
        Matrix to convert DSS components back to sensor space.
    pwr0: array
        Power per component (raw).
    pwr1: array
        Power per component (averaged).

    """
    # if demean: # remove weighted mean
    #   X = demean(X, weights)

    # weighted mean over trials (--> bias function for DSS)
    xx, ww = mean_over_trials(X, weights)

    # covariance of raw and biased X
    c0, nc0 = tscov(X, None, weights)
    c1, nc1 = tscov(xx, None, ww)
    c0 /= nc0
    c1 /= nc1

    todss, fromdss, pwr0, pwr1 = dss0(c0, c1, keep1, keep2)

    return todss, fromdss, pwr0, pwr1


def dss0(c0, c1, keep1=None, keep2=1e-9, return_unmixing=True):
    """DSS base function.

    This function allows specifying arbitrary bias functions (as compared to
    the function:`dss1`, which forces the bias to be the mean over trials).

    Parameters
    ----------
    c0: array, shape=(n_chans, n_chans)
        Baseline covariance.
    c1: array, shape=(n_chans, n_chans)
        Biased covariance.
    keep1: int | None
        Number of PCs to retain (default=None, which keeps all).
    keep2: float
        Ignore PCs smaller than keep2 (default=1e-9).
    return_unmixing : bool
        If True (default), return the unmixing matrix.

    Returns
    -------
    todss: array, shape=(n_chans, n_dss_components)
        Matrix to convert X to normalized DSS components.
    fromdss : array, shape=(n_dss_components, n_chans)
        Matrix to transform back to original space. Only returned if
        ``return_unmixing`` is True.
    pwr0: array
        Power per component (baseline).
    pwr1: array
        Power per component (biased).

    Notes
    -----
    The data mean is NOT removed prior to processing.

    """
    # Check size and squareness
    assert c0.shape == c1.shape == (c0.shape[0], c0.shape[0]), \
        "c0 and c1 should have the same size, and be square"

    # Check for NaN or INF
    assert not (np.any(np.isnan(c0)) or np.any(np.isinf(c0))), "NaN or INF in c0"
    assert not (np.any(np.isnan(c1)) or np.any(np.isinf(c1))), "NaN or INF in c1"

    # derive PCA and whitening matrix from unbiased covariance
    eigvec0, eigval0 = pca(c0, max_comps=keep1, thresh=keep2)

    # apply whitening and PCA matrices to the biased covariance
    # (== covariance of bias whitened data)
    W = np.diag(np.sqrt(1. / eigval0))  # diagonal of whitening matrix

    # c1 is projected into whitened PCA space of data channels
    c2 = (eigvec0 @ W).T @ c1 @ (eigvec0 @ W)

    # proj. matrix from whitened data space to a space maximizing bias
    eigvec2, eigval2 = pca(c2, max_comps=keep1, thresh=keep2)

    # DSS matrix (raw data to normalized DSS)
    todss = eigvec0 @ W @ eigvec2

    # Normalise DSS matrix
    N = np.sqrt(np.diag(todss.T @ c0 @ todss))
    todss /= N

    pwr0 = np.sqrt(np.sum((c0 @ todss) ** 2, axis=0))
    pwr1 = np.sqrt(np.sum((c1 @ todss) ** 2, axis=0))

    # Return data
    # next line equiv. to: np.array([np.dot(todss, ep) for ep in data])
    # dss_data = np.einsum('ij,hjk->hik', todss, data)

    if return_unmixing:
        fromdss = linalg.pinv(todss)
        return todss, fromdss, pwr0, pwr1
    else:
        return todss, pwr0, pwr1


def dss_line(X, fline, sfreq, nremove=1, nfft=1024, nkeep=None, blocksize=None,
             show=False):
    """Apply DSS to remove power line artifacts.

    Implements the ZapLine algorithm described in [1]_.

    Parameters
    ----------
    X : data, shape=(n_samples, n_chans, n_trials)
        Input data.
    fline : float
        Line frequency (normalized to sfreq, if ``sfreq`` == 1).
    sfreq : float
        Sampling frequency (default=1, which assymes ``fline`` is normalised).
    nremove : int
        Number of line noise components to remove (default=1).
    nfft : int
        FFT size (default=1024).
    nkeep : int
        Number of components to keep in DSS (default=None).
    blocksize : int
        If not None (default), covariance is computed on blocks of
        ``blocksize`` samples. This may improve performance for large datasets.
    show: bool
        If True, show DSS results (default=False).

    Returns
    -------
    y : array, shape=(n_samples, n_chans, n_trials)
        Denoised data.
    artifact : array, shape=(n_samples, n_chans, n_trials)
        Artifact

    Examples
    --------
    Apply to X, assuming line frequency=50Hz and sampling rate=1000Hz, plot
    results:
    >>> dss_line(X, 50/1000)

    Removing 4 line-dominated components:
    >>> dss_line(X, 50/1000, 4)

    Truncating PCs beyond the 30th to avoid overfitting:
    >>> dss_line(X, 50/1000, 4, nkeep=30);

    Return cleaned data in y, noise in yy, do not plot:
    >>> [y, artifact] = dss_line(X, 60/1000)

    References
    ----------
    .. [1] de Cheveigné, A. (2019). ZapLine: A simple and effective method to remove
       power line artifacts. NeuroImage, 116356.
       https://doi.org/10.1016/j.neuroimage.2019.116356

    """
    if X.shape[0] < nfft:
        print(f"Reducing nfft to {X.shape[0]}")
        nfft = X.shape[0]
    n_samples, n_chans, _ = theshapeof(X)
    if blocksize is None:
        blocksize = n_samples

    # Recentre data
    X = demean(X, inplace=True)

    # Cancel line_frequency and harmonics + light lowpass
    X_filt = smooth(X, sfreq / fline)

    # X - X_filt results in the artifact plus some residual biological signal
    X_noise = X - X_filt

    # Reduce dimensionality to avoid overfitting
    if nkeep is not None:
        cov_X_res = tscov(X_noise)[0]
        V, _ = pca(cov_X_res, nkeep)
        X_noise_pca = X_noise @ V
    else:
        X_noise_pca = X_noise.copy()
        nkeep = n_chans

    # Compute blockwise covariances of raw and biased data
    n_harm = np.floor((sfreq / 2) / fline).astype(int)
    c0 = np.zeros((nkeep, nkeep))
    c1 = np.zeros((nkeep, nkeep))
    for X_block in sliding_window_view(X_noise_pca, (blocksize, nkeep),
                                       axis=(0, 1))[::blocksize, 0]:
        # if n_trials>1, reshape to (n_samples, nkeep, n_trials)
        if X_block.ndim == 3:
            X_block = X_block.transpose(1, 2, 0)

        # bias data
        c0 += tscov(X_block)[0]
        c1 += tscov(gaussfilt(X_block, sfreq, fline, fwhm=1, n_harm=n_harm))[0]

    # DSS to isolate line components from residual
    todss, _, pwr0, pwr1 = dss0(c0, c1)

    if show:
        import matplotlib.pyplot as plt
        plt.plot(pwr1 / pwr0, ".-")
        plt.xlabel("component")
        plt.ylabel("score")
        plt.title("DSS to enhance line frequencies")
        plt.show()

    # Remove line components from X_noise
    idx_remove = np.arange(nremove)
    X_artifact = matmul3d(X_noise_pca, todss[:, idx_remove])
    X_res = tsr(X_noise, X_artifact)[0]  # project them out

    # reconstruct clean signal
    y = X_filt + X_res

    # Power of components
    p = wpwr(X - y)[0] / wpwr(X)[0]
    print(f"Power of components removed by DSS: {p:.2f}")
    # return the reconstructed clean signal, and the artifact
    return y, X - y


def dss_line_iter(data, fline, sfreq, win_sz=10, spot_sz=2.5,
                  nfft=512, show=False, prefix="dss_iter", n_iter_max=100):
    """Remove power line artifact iteratively.

    This method applies dss_line() until the artifact has been smoothed out
    from the spectrum.

    Parameters
    ----------
    data : data, shape=(n_samples, n_chans, n_trials)
        Input data.
    fline : float
        Line frequency.
    sfreq : float
        Sampling frequency.
    win_sz : float
        Half of the width of the window around the target frequency used to fit
        the polynomial (default=10).
    spot_sz : float
        Half of the width of the window around the target frequency used to
        remove the peak and interpolate (default=2.5).
    nfft : int
        FFT size for the internal PSD calculation (default=512).
    show: bool
        Produce a visual output of each iteration (default=False).
    prefix : str
        Path and first part of the visualisation output file
        "{prefix}_{iteration number}.png" (default="dss_iter").
    n_iter_max : int
        Maximum number of iterations (default=100).

    Returns
    -------
    data : array, shape=(n_samples, n_chans, n_trials)
        Denoised data.
    iterations : int
        Number of iterations.
    """

    def nan_basic_interp(array):
        """Nan interpolation."""
        nans, ix = np.isnan(array), lambda x: x.nonzero()[0]
        array[nans] = np.interp(ix(nans), ix(~nans), array[~nans])
        return array

    freq_rn = [fline - win_sz, fline + win_sz]
    freq_sp = [fline - spot_sz, fline + spot_sz]
    freq, psd = welch(data, fs=sfreq, nfft=nfft, axis=0)

    freq_rn_ix = np.logical_and(freq >= freq_rn[0], freq <= freq_rn[1])
    freq_used = freq[freq_rn_ix]
    freq_sp_ix = np.logical_and(freq_used >= freq_sp[0],
                                freq_used <= freq_sp[1])

    if psd.ndim == 3:
        mean_psd = np.mean(psd, axis=(1, 2))[freq_rn_ix]
    elif psd.ndim == 2:
        mean_psd = np.mean(psd, axis=(1))[freq_rn_ix]

    mean_psd_wospot = mean_psd.copy()
    mean_psd_wospot[freq_sp_ix] = np.nan
    mean_psd_tf = nan_basic_interp(mean_psd_wospot)
    pf = np.polyfit(freq_used, mean_psd_tf, 3)
    p = np.poly1d(pf)
    clean_fit_line = p(freq_used)

    aggr_resid = []
    iterations = 0
    while iterations < n_iter_max:
        data, _ = dss_line(data, fline, sfreq, nfft=nfft, nremove=1)
        freq, psd = welch(data, fs=sfreq, nfft=nfft, axis=0)
        if psd.ndim == 3:
            mean_psd = np.mean(psd, axis=(1, 2))[freq_rn_ix]
        elif psd.ndim == 2:
            mean_psd = np.mean(psd, axis=(1))[freq_rn_ix]

        residuals = mean_psd - clean_fit_line
        mean_score = np.mean(residuals[freq_sp_ix])
        aggr_resid.append(mean_score)

        print(f"Iteration {iterations} score: {mean_score}")

        if show:
            import matplotlib.pyplot as plt
            f, ax = plt.subplots(2, 2, figsize=(12, 6), facecolor="white")

            if psd.ndim == 3:
                mean_sens = np.mean(psd, axis=2)
            elif psd.ndim == 2:
                mean_sens = psd

            y = mean_sens[freq_rn_ix]
            ax.flat[0].plot(freq_used, y)
            ax.flat[0].set_title("Mean PSD across trials")

            ax.flat[1].plot(freq_used, mean_psd_tf, c="gray")
            ax.flat[1].plot(freq_used, mean_psd, c="blue")
            ax.flat[1].plot(freq_used, clean_fit_line, c="red")
            ax.flat[1].set_title("Mean PSD across trials and sensors")

            tf_ix = np.where(freq_used <= fline)[0][-1]
            ax.flat[2].plot(residuals, freq_used)
            color = "green"
            if mean_score <= 0:
                color = "red"
            ax.flat[2].scatter(residuals[tf_ix], freq_used[tf_ix], c=color)
            ax.flat[2].set_title("Residuals")

            ax.flat[3].plot(np.arange(iterations + 1), aggr_resid, marker="o")
            ax.flat[3].set_title("Iterations")

            plt.tight_layout()
            plt.savefig(f"{prefix}_{iterations:03}.png")
            plt.close("all")

        if mean_score <= 0:
            break

        iterations += 1

    if iterations == n_iter_max:
        raise RuntimeError("Could not converge. Consider increasing the "
                           "maximum number of iterations")

    return data, iterations



"""TRCA utils."""
import numpy as np
from scipy import stats
from scipy.signal import cheb1ord, cheby1, filtfilt


def round_half_up(num, decimals=0):
    """Round half up round the last decimal of the number.

    The rules are:
    from 0 to 4 rounds down
    from 5 to 9 rounds up

    Parameters
    ----------
    num : float
        Number to round
    decimals : number of decimals

    Returns
    -------
    num rounded
    """
    multiplier = 10 ** decimals
    return int(np.floor(num * multiplier + 0.5) / multiplier)


def normfit(data, ci=0.95):
    """Compute the mean, std and confidence interval for them.

    Parameters
    ----------
    data : array, shape=()
        Input data.
    ci : float
        Confidence interval (default=0.95).

    Returns
    -------
    m : float
        Mean.
    sigma : float
        Standard deviation
    [m - h, m + h] : list
        Confidence interval of the mean.
    [sigmaCI_lower, sigmaCI_upper] : list
        Confidence interval of the std.
    """
    arr = 1.0 * np.array(data)
    num = len(arr)
    avg, std_err = np.mean(arr), stats.sem(arr)
    h_int = std_err * stats.t.ppf((1 + ci) / 2., num - 1)
    var = np.var(data, ddof=1)
    var_ci_upper = var * (num - 1) / stats.chi2.ppf((1 - ci) / 2, num - 1)
    var_ci_lower = var * (num - 1) / stats.chi2.ppf(1 - (1 - ci) / 2, num - 1)
    sigma = np.sqrt(var)
    sigma_ci_lower = np.sqrt(var_ci_lower)
    sigma_ci_upper = np.sqrt(var_ci_upper)

    return avg, sigma, [avg - h_int, avg +
                        h_int], [sigma_ci_lower, sigma_ci_upper]


def itr(n, p, t):
    """Compute information transfer rate (ITR).

    Definition in [1]_.

    Parameters
    ----------
    n : int
        Number of targets.
    p : float
        Target identification accuracy (0 <= p <= 1).
    t : float
        Average time for a selection (s).

    Returns
    -------
    itr : float
        Information transfer rate [bits/min]

    References
    ----------
    .. [1] M. Cheng, X. Gao, S. Gao, and D. Xu,
        "Design and Implementation of a Brain-Computer Interface With High
        Transfer Rates", IEEE Trans. Biomed. Eng. 49, 1181-1186, 2002.

    """
    itr = 0

    if (p < 0 or 1 < p):
        raise ValueError("Accuracy need to be between 0 and 1.")
    elif (p < 1 / n):
        itr = 0
        raise ValueError("ITR might be incorrect because accuracy < chance")
    elif (p == 1):
        itr = np.log2(n) * 60 / t
    else:
        itr = (np.log2(n) + p * np.log2(p) + (1 - p) *
               np.log2((1 - p) / (n - 1))) * 60 / t

    return itr


def bandpass(eeg, sfreq, Wp, Ws):
    """Filter bank design for decomposing EEG data into sub-band components.

    Parameters
    ----------
    eeg : np.array, shape=(n_samples, n_chans[, n_trials])
        Training data.
    sfreq : int
        Sampling frequency of the data.
    Wp : 2-tuple
        Passband for Chebyshev filter.
    Ws : 2-tuple
        Stopband for Chebyshev filter.

    Returns
    -------
    y: np.array, shape=(n_trials, n_chans, n_samples)
        Sub-band components decomposed by a filter bank.

    See Also
    --------
    scipy.signal.cheb1ord :
        Chebyshev type I filter order selection.

    """
    # Chebyshev type I filter order selection.
    N, Wn = cheb1ord(Wp, Ws, 3, 40, fs=sfreq)

    # Chebyshev type I filter design
    B, A = cheby1(N, 0.5, Wn, btype="bandpass", fs=sfreq)

    # the arguments 'axis=0, padtype='odd', padlen=3*(max(len(B),len(A))-1)'
    # correspond to Matlab filtfilt : https://dsp.stackexchange.com/a/47945
    y = filtfilt(B, A, eeg, axis=0, padtype="odd",
                 padlen=3 * (max(len(B), len(A)) - 1))
    return y


def schaefer_strimmer_cov(X):
    r"""Schaefer-Strimmer covariance estimator.

    Shrinkage estimator described in [1]_:

    .. math:: \hat{\Sigma} = (1 - \gamma)\Sigma_{scm} + \gamma T

    where :math:`T` is the diagonal target matrix:

    .. math:: T_{i,j} = \{ \Sigma_{scm}^{ii} \text{if} i = j,
         0 \text{otherwise} \}

    Note that the optimal :math:`\gamma` is estimated by the authors' method.

    Parameters
    ----------
    X: array, shape=(n_chans, n_samples)
        Signal matrix.

    Returns
    -------
    cov: array, shape=(n_chans, n_chans)
        Schaefer-Strimmer shrinkage covariance matrix.

    References
    ----------
    .. [1] Schafer, J., and K. Strimmer. 2005. A shrinkage approach to
       large-scale covariance estimation and implications for functional
       genomics. Statist. Appl. Genet. Mol. Biol. 4:32.
    """
    ns = X.shape[1]
    C_scm = np.cov(X, ddof=0)
    X_c = X - np.tile(X.mean(axis=1), [ns, 1]).T

    # Compute optimal gamma, the weigthing between SCM and srinkage estimator
    R = ns / (ns - 1.0) * np.corrcoef(X)
    var_R = (X_c ** 2).dot((X_c ** 2).T) - 2 * C_scm * X_c.dot(X_c.T)
    var_R += ns * C_scm ** 2

    var_R = ns / ((ns - 1) ** 3 * np.outer(X.var(1), X.var(1))) * var_R
    R -= np.diag(np.diag(R))
    var_R -= np.diag(np.diag(var_R))
    gamma = max(0, min(1, var_R.sum() / (R ** 2).sum()))

    cov = (1. - gamma) * (ns / (ns - 1.)) * C_scm
    cov += gamma * (ns / (ns - 1.)) * np.diag(np.diag(C_scm))

    return cov




"""Real-time phase and amplitude estimation using resonant oscillators.

Notes
-----
While the original paper [1]_ only describes the algorithm for a single
channel, this implementation can handle multiple channels. However, the
algorithm is not optimized for this purpose. Therefore, this code is likely to
be slow for large input arrays (n_channels >> 10), since an individual
oscillator is instantiated for each channel.

.. [1] Rosenblum, M., Pikovsky, A., Kühn, A.A. et al. Real-time estimation of
       phase and amplitude with application to neural data. Sci Rep 11, 18037
       (2021). https://doi.org/10.1038/s41598-021-97560-5
"""
import numpy as np
from scipy.fftpack import fft, fftshift, ifft, ifftshift, next_fast_len
from scipy.signal import butter, freqz

from .utils.buffer import Buffer


class Device:
    """Measurement device.

    This class implements a measurement device, as describes by Rosenblum et al.

    Parameters
    ----------
    om0 : float
        Oscillator frequency (estimation).
    dt : float
        Sampling interval.
    damping : float
        Damping parameter.
    type : str in {"integrator", "oscillator"}
        Type of the device.
    mu : float
        Integrator parameter (only for type="integrator"). Should be >> target
        frequency of interest.

    """

    def __init__(self, om0, dt, damping, type, mu=500):
        self.om0 = om0
        self.dt = dt
        self.damping = damping
        self.type = type
        self.init_coefs(om0, dt, damping)

        # For the integrator
        if type == "integrator":
            self.mu = mu
            self.edelmu = np.exp(-self.dt / self.mu)

        self.state = dict(x=0, y=0)


    def init_coefs(self, om0, dt, damping):
        """Initialize coefficients for solving oscillator's equations."""
        self.c1, self.c2, self.c3, self.enuDel, self.ealDel, self.eta = \
            init_coefs(om0, dt, damping)
        self.om0 = om0

    def step(self, sprev, s, snew):
        """Perform one step of the oscillator equations.

        Parameters
        ----------
        x : float
            State variable x.
        y : float
            State variable y.
        sprev : float
            Previous measurement.
        s : float
            Current measurement.
        snew : float
            New measurement.

        Returns
        -------
        x : float
            Updated state variable x.
        y : float
            Updated state variable y.
        """
        if self.type == "oscillator":
            x = self.state["x"]
            y = self.state["y"]
            x, y = one_step_oscillator(x, y, self.damping, self.eta, self.enuDel,
                                       self.ealDel, self.c1, self.c2, self.c3,
                                       sprev, s, snew)
            self.state["x"] = x
            self.state["y"] = y

        elif self.type == "integrator":
            x = self.state["x"]
            x = one_step_integrator(x, self.edelmu, self.mu, self.dt,
                                    sprev, s, snew)
            self.state["x"] = x

        return self


class NonResOscillator:
    """Real-time measurement of phase and amplitude using non-resonant oscillator.

    This estimator relies on the resonance effect. The measuring “device”
    consists of two linear damped oscillators. The oscillators' frequency is
    much larger than the frequency of the signal, i.e., the system is far from
    resonance. We choose the damping parameters to ensure that (i) the phase of
    the first linear oscillator equals that of the input and that (ii)
    amplitude of the second one and the input relate by a known constant
    multiplicator. The technique yields both phase and amplitude of the input
    signal.

    This estimator includes an automated frequency-tuning algorithm to adjust
    to the a priori unknown signal frequency.

    For a detailed description, refer to [1]_.

    Parameters
    ----------
    fs : float
        Sampling frequency in Hz.
    nu : float
        Rough estimate of the main frequency of interest.

    References
    ----------
    .. [1] Rosenblum, M., Pikovsky, A., Kühn, A.A. et al. Real-time estimation
        of phase and amplitude with application to neural data. Sci Rep 11,
        18037 (2021). https://doi.org/10.1038/s41598-021-97560-5
    """

    def __init__(self, fs=250, nu=1.1, alpha_a=6.0, alpha_p=0.2, update_factor=5):

        # Parameters of the measurement "devices"
        self.dt = 1 / fs  # Sampling interval
        self.nu = nu  # Rough estimate of the tremor frequency
        self.om0 = 5 * nu  # Oscillator frequency (estimation)
        self.alpha_a = alpha_a  # Damping parameter for the "amplitude device"
        self.gamma_a = alpha_a / 2
        self.alpha_p = alpha_p  # Damping parameter for the "phase device"
        self.gamma_p = alpha_p / 2
        self.factor = np.sqrt((self.om0 ** 2 - nu ** 2) ** 2 + (self.alpha_a * nu) ** 2)

        # Update parameters, and precomputed quantities
        self.memory = round(2 * np.pi / self.om0 / self.dt)
        self.update_point = 2 * self.memory
        self.update_step = round(self.memory / update_factor)
        self._tbuf = np.arange(self.memory) * self.dt
        self._Sx = np.sum(self._tbuf)
        self._denom = self.memory * np.sum(self._tbuf ** 2) - self._Sx ** 2

        # Buffer to store past input values
        self.n_channels = None
        self.buffer = None

    def _set_devices(self, n_channels):
        # Set up the phase and amplitude "devices"
        self.adevice = [Device(self.om0, self.dt, self.gamma_a, "oscillator")
                        for _ in range(n_channels)]
        self.pdevice = [Device(self.om0, self.dt, self.gamma_p, "oscillator")
                        for _ in range(n_channels)]


    def transform(self, X):
        """Transform the input signal into phase and amplitude estimates.

        Parameters
        ----------
        X : ndarray, shape=(n_samples, n_channels)
            The input signal to be transformed.

        Returns
        -------
        phase : ndarray, shape=(n_samples, n_channels)
            Current phase estimate.
        ampl : ndarray, shape=(n_samples, n_channels)
            Current amplitude estimate.
        """
        if self.buffer is None:
            self.n_channels = 1 if X.ndim < 2 else X.shape[1]
            self.buffer = Buffer(self.memory, self.n_channels)
            self.phase_buffer = Buffer(self.memory, self.n_channels)
            self._set_devices(self.n_channels)

        n_samples = X.shape[0]
        phase = np.zeros((n_samples, self.n_channels))
        amp = np.zeros((n_samples, self.n_channels))

        for k in range(n_samples):
            self.buffer.push(X[k])

            if self.buffer.counter < 3:
                continue # Skip the first two samples

            # Amplitude estimation
            spp, sp, s = self.buffer.view(3)[:, 0]

            for ch in range(self.n_channels):
                self.adevice[ch].step(spp, sp, s)
                z = self.adevice[ch].state["y"] / self.nu
                amp[k, ch] = self.factor * \
                    np.sqrt(z ** 2 + self.adevice[ch].state["x"] ** 2)

                # Phase estimation
                self.pdevice[ch].step(spp, sp, s)
                z = self.pdevice[ch].state["y"] / self.nu
                phase[k, ch] = np.arctan2(-z, self.pdevice[ch].state["x"])

            self.phase_buffer.push(phase[k])

            if k > self.update_point:
                tmpphase = self.phase_buffer.view(self.memory)
                for ch in range(self.n_channels):
                    tmp = np.unwrap(tmpphase[:, ch])
                    self.nu = (self.memory * np.sum(self._tbuf * tmp) -
                               self._Sx * np.sum(tmp)) / self._denom
                self.update_point += self.update_step

        return phase, amp


class ResOscillator:
    """Real-time measurement of phase and amplitude using a resonant oscillator.

    In this version, the corresponding “device” consists of a linear oscillator
    in resonance with the measured signal and of an integrating unit. It also
    provides both phase and amplitude of the input signal. The method exploits
    the known relation between the resonant oscillator's phase and amplitude
    and those of the input. Additionally, the resonant oscillator acts as a
    bandpass filter for experimental data.

    This filter also includes a frequency-adaptation algorithm, and removes
    low-frequency trend (baseline fluctuations). For a detailed description,
    refer to [1]_.

    Parameters
    ----------
    npt : int
        Number of points to be measured and processed
    fs : float
        Sampling frequency in Hz.
    nu : float
        Rough estimate of the tremor frequency.
    freq_adaptation : bool
        Whether to use the frequency adaptation algorithm (default=True).

    Returns
    -------
    Dphase : numpy.ndarray
        Array containing the phase values.
    Dampl : numpy.ndarray
        Array containing the amplitude values.

    References
    ----------
    .. [1] Rosenblum, M., Pikovsky, A., Kühn, A.A. et al. Real-time estimation
       of phase and amplitude with application to neural data. Sci Rep 11, 18037
       (2021). https://doi.org/10.1038/s41598-021-97560-5
    """

    def __init__(self, fs=1000, nu=4.5, update_factor=5, freq_adaptation=True,
                 assume_centered=False):

        # Parameters of the measurement "device"
        self.dt = 1 / fs  # Sampling interval
        self.om0 = nu  # Angular frequency
        self.alpha = 0.3 * self.om0
        self.gamma = self.alpha / 2

        # Parameters of adaptation algorithm
        nperiods = 1  # Number of previous periods for frequency correction
        npt_period = round(2 * np.pi / self.om0 / self.dt)  # Number of points per period
        self.memory = nperiods * npt_period  # M points for frequency correction buffer
        self.update_factor = update_factor  # Number of frequency updates per period
        self.update_step = round(npt_period / self.update_factor)
        self.updatepoint = 2 * self.memory

        # Precomputed quantities for linear fit for frequency adaptation
        self._tbuf = np.arange(1, self.memory + 1) * self.dt
        self._Sx = np.sum(self._tbuf)
        self._denom = self.memory * np.sum(self._tbuf ** 2) - self._Sx ** 2

        # Buffer to store past input values
        self.n_channels = None
        self.buffer = None
        self.runav = 0.  # Initial guess for the dc-component
        self.freq_adaptation = freq_adaptation
        self.assume_centered = assume_centered

    def _set_devices(self, n_channels):
        # Set up the phase and amplitude "devices"
        self.osc = [Device(self.om0, self.dt, self.gamma, "oscillator")
                    for _ in range(n_channels)]
        self.int = [Device(self.om0, self.dt, self.gamma, "integrator")
                    for _ in range(n_channels)]

    def transform(self, X):
        """Transform the input signal into phase and amplitude estimates.

        Parameters
        ----------
        X : ndarray, shape (n_samples, n_channels)
            The input signal to be transformed.

        Returns
        -------
        phase : float
            Current phase estimate.
        ampl : float
            Current amplitude estimate.
        """
        if self.buffer is None:
            self.n_channels = 1 if X.ndim < 2 else X.shape[1]
            self.buffer = Buffer(self.memory, self.n_channels)
            self.phase_buffer = Buffer(self.memory, self.n_channels)
            self._demean = np.zeros((3, self.n_channels)) # past demeaned signal values
            self._oscbuf = np.zeros((3, self.n_channels)) # past integrator values
            self._set_devices(self.n_channels)

        n_samples = X.shape[0]
        phase = np.zeros((n_samples, self.n_channels))
        ampl = np.zeros((n_samples, self.n_channels))

        for k in range(n_samples):
            self.buffer.push(X[k])
            self._demean = np.roll(self._demean, -1, axis=0)
            self._demean[-1] = self.buffer.view(1) - self.runav  # Baseline correction

            if self.buffer.counter < 3:
                continue # Skip the first two samples

            for ch in range(self.n_channels):
                # Perform one step of the oscillator equations
                self.osc[ch].step(self._demean[-3, ch],
                                  self._demean[-2, ch],
                                  self._demean[-1, ch])

                self._oscbuf[:, ch] = np.roll(self._oscbuf[:, ch], -1, axis=0)
                self._oscbuf[-1, ch] = self.osc[ch].state["y"]

                self.int[ch].step(self._oscbuf[-3, ch],
                                  self._oscbuf[-2, ch],
                                  self._oscbuf[-1, ch])

                # New phase and amplitude values
                v = self.int[ch].mu * self.int[ch].state["x"] * self.osc[ch].om0

                phase[k, ch] = np.arctan2(v, np.real(self.osc[ch].state["y"]))
                ampl[k, ch] = self.alpha * np.sqrt(self.osc[ch].state["y"] ** 2 + v ** 2)

            self.phase_buffer.push(phase[k])

            if self.freq_adaptation and k > self.updatepoint:
                tmpphase = self.phase_buffer.view(self.memory)
                # Recompute integration parameters for the new frequency
                for ch in range(self.n_channels):
                    tmp = np.unwrap(tmpphase[:, ch])
                    # Frequency via linear fit of the phases in the buffer
                    om0 = (self.memory * np.sum(self._tbuf * tmp) -
                        self._Sx * np.sum(tmp)) / self._denom
                    self.osc[ch].init_coefs(om0, self.dt, self.gamma)

                # Update running average
                if self.assume_centered is False:
                    self.runav = np.mean(self.buffer.view(self.memory), axis=0,
                                         keepdims=True)
                self.updatepoint += self.update_step  # Point for the next update

        return phase, ampl


def locking_based_phase(s, dt, npt):
    """Compute the locking-based phase.

    This technique exploits the ideas from the synchronization theory. It is
    well-known that a force s(t) acting on a limit-cycle oscillator can entrain
    it. It means that the oscillator's frequency becomes equal to that of the
    force, and their phases differ by a constant. Thus, the phase of the locked
    limit-cycle oscillator will correspond to the phase of the signal. For our
    purposes, it is helpful to use the simplest oscillator model, the so-called
    phase oscillator. To ensure the phase-locking to the force, we have to
    adjust the oscillator's frequency to the signal's frequency. We assume that
    we do not know the latter a priori, but can only roughly estimate it. We
    propose a simple approach that starts with this estimate and automatically
    tunes the “device's” frequency to ensure the locking and thus provides the
    instantaneous phase.

    Note that the amplitude is not determined with this approach. Technically,
    the algorithm reduces to solving differential equation incorporating
    measured data given at discrete time points; for details, see [1]_.

    Parameters
    ----------
    s: array-like
        Input signal.
    dt: float
        Time step.
    npt: int
        Number of points.

    Returns
    -------
    lb_phase: ndarray
        Locking-based phase.

    References
    ----------
    .. [1] Rosenblum, M., Pikovsky, A., Kühn, A.A. et al. Real-time estimation
       of phase and amplitude with application to neural data. Sci Rep 11, 18037
       (2021). https://doi.org/10.1038/s41598-021-97560-5
    """
    lb_phase = np.zeros(npt)

    # Parameters of the measurement device
    epsilon = 0.8  # Amplitude of the oscillator
    K = 0.5  # Frequency adaptation coefficient
    n_rk_steps = 5  # Number of Runge-Kutta steps per time step

    # Parameters of the frequency adaptation algorithm
    nu_est = 1.1  # initial guess for the frequency
    update_factor = 20  # number of frequency updates per period

    # Buffer for frequency estimation
    npbuf = round(2 * np.pi / nu_est / dt) # number of samples for frequency estimation
    update_point = 2 * npbuf
    update_step = round(npbuf / update_factor)
    tbuf = np.arange(npbuf) * dt
    sx = np.sum(tbuf)
    denom = npbuf * np.sum(tbuf ** 2) - sx ** 2

    for k in range(2, npt):
        lb_phase[k] = rk(lb_phase[k - 1], dt, n_rk_steps, nu_est, epsilon,
                         s[k], s[k - 1], s[k])

        if k > update_point:
            buffer = lb_phase[k - npbuf:k]
            nu_quasi = (npbuf * np.sum(tbuf * buffer) - sx * np.sum(buffer)) / denom
            nu_est = nu_est + K * (nu_quasi - nu_est)
            update_point += update_step

    return lb_phase


def rk(phi, dt, n_steps, omega, epsilon, s_prev, s, s_new):
    """Runge-Kutta method for phase calculation.

    Parameters
    ----------
    phi : float
        Initial phase value.
    dt : float
        Time step size.
    n_steps : int
        Number of steps to iterate.
    omega : float
        Angular frequency.
    epsilon : float
        Amplitude of the oscillation.
    s_prev : float
        Previous phase value.
    s : float
        Current phase value.
    s_new : float
        New phase value.

    Returns
    -------
    phi : float
        The calculated phase value.

    """
    h = dt / n_steps
    b = (s_new - s_prev) / dt / 2
    c = (s_prev - 2 * s + s_new) / dt ** 2
    hh = h / 2.
    h6 = h / 6.
    t = 0

    for _ in range(n_steps):
        th = t + hh
        dy0 = omega - epsilon * (s + b * t + c * t ** 2) * np.sin(phi)
        phit = phi + hh * dy0
        dyt = omega - epsilon * (s + b * th + c * th ** 2) * np.sin(phit)
        phit = phi + hh * dyt
        dym = omega - epsilon * (s + b * th + c * th ** 2) * np.sin(phit)
        phit = phi + h * dym
        dym = dym + dyt
        t = t + h
        dyt = omega - epsilon * (s + b * t + c * t ** 2) * np.sin(phit)
        phi = phi + h6 * (dy0 + 2 * dym + dyt)

    return phi


def init_coefs(om0, dt, alpha):
    """Compute coefficients for solving oscillator's equations.

    Parameters
    ----------
    om0 : float
        Oscillator frequency (estimation).
    dt : float
        Sampling interval.
    alpha : float
        Half of the damping.

    Returns
    -------
    C1 : float
        Coefficient C1.
    C2 : float
        Coefficient C2.
    C3 : float
        Coefficient C3.
    eetadel : complex
        Exponential term for amplitude device.
    ealdel : float
        Exponential term for amplitude device.
    eta : float
        Square root of the difference of oscillator frequency squared and
        damping squared.

    """
    # alpha is the half of the damping: x'' + 2 * alpha * x' + om0^2 * x = input
    eta2 = om0**2 - alpha**2
    eta2 = complex(eta2) if eta2 < 0 else eta2
    eta = np.sqrt(eta2)  # replicate Matlab behavior
    eetadel = np.exp(1j * eta * dt)
    a = 1. / eetadel
    ealdel = np.exp(alpha * dt)

    I1 = 1j * (a - 1) / eta
    I2 = (a * (1 + 1j * eta * dt) - 1) / eta2
    I3 = (a * (dt * eta * (2 + 1j * dt * eta) - 2 * 1j) + 2 * 1j) / eta2 / eta

    C1 = (I3 - I2 * dt) / (2 * dt**2 * ealdel)
    C2 = I1 - I3 / dt**2
    C3 = ealdel * (I2 * dt + I3) / (2 * dt**2)

    return C1, C2, C3, eetadel, ealdel, eta


def one_step_oscillator(x, xd, alpha, eta, eeta_del, eal_del, C1, C2, C3, spp, sp, s):
        """Perform one step of the oscillator equations.

        Parameters
        ----------
        x : float
            State variable x.
        xd : float
            Derivative of state variable x.
        alpha : float
            Damping parameter.
        eta : float
            Square root of the difference of oscillator frequency squared and
            damping squared.
        eeta_del : complex
            Exponential term for time step.
        eal_del : float
            Exponential term for time step.
        C1 : float
            Coefficient C1.
        C2 : float
            Coefficient C2.
        C3 : float
            Coefficient C3.
        spp : float
            Second orevious measurement.
        sp : float
            Previous measurement.
        s : float
            Current measurement.

        Returns
        -------
        x : float
            Updated state variable x.
        xd : float
            Updated derivative of state variable x.
        """
        A = x - 1j * (xd + alpha * x + C1 * spp + C2 * sp + C3 * s) / eta
        d = A * eeta_del
        y = np.real(d)
        yd = - eta * np.imag(d)
        x = y / eal_del
        xd = (yd - alpha * y) / eal_del

        return x, xd


def one_step_integrator(z, edelmu, mu, dt, spp, sp, s):
    """Perform one step of the integrator in the oscillator equations.

    Parameters
    ----------
    z : float
        State variable z.
    edelmu : float
        Exponential term for integrator.
    mu : float
        Integrator parameter.
    dt : float
        Sampling interval.
    spp : float
        Second previous measurement.
    sp : float
        Previous measurement.
    s : float
        Current measurement.

    Returns
    -------
    z : float
        Updated state variable z.
    """
    a = sp
    b = (s - spp) / 2 / dt
    c = (spp - 2 * sp + s) / 2 / dt ** 2
    d = -a + b * mu - 2 * c * (mu ** 2)
    C0 = z + d
    z = C0 * edelmu - d + b * dt - 2 * c * mu * dt + c * dt ** 2
    return z


class ECHT:
    """Endpoint Corrected Hilbert Transform (ECHT).

    See [1]_ for details.

    Parameters
    ----------
    X : ndarray, shape=(n_samples, n_channels)
        Time domain signal.
    l_freq : float | None
        Low-cutoff frequency of a bandpass causal filter. If None, the data is
        only low-passed.
    h_freq : float | None
        High-cutoff frequency of a bandpass causal filter. If None, the data is
        only high-passed.
    sfreq : float
        Sampling rate of time domain signal.
    n_fft : int, optional
        Length of analytic signal. If None, it defaults to the length of X.
    filt_order : int, optional
        Order of the filter. Default is 2.

    Notes
    -----
    One common implementation of the Hilbert Transform uses a DFT (aka FFT)
    as part of its computation. Inherent to the DFT is the assumption that
    a finite sample of a signal is replicated infinitely in time, effectively
    abutting the end of a sample with its replicated start. If the start and
    end of the sample are not continuous with each other, distortions are
    introduced by the DFT. Echt effectively smooths out this 'discontinuity'
    by selectively deforming the start of the sample. It is hence most suited
    for real-time applications in which the point/s of interest is/are the
    most recent one/s (i.e. last) in the sample window.

    We found that a filter bandwidth (BW=h_freq-l_freq) of up to half the
    signal's central frequency works well.

    References
    ----------
    .. [1] Schreglmann, S. R., Wang, D., Peach, R. L., Li, J., Zhang, X.,
        Latorre, A., ... & Grossman, N. (2021). Non-invasive suppression of
        essential tremor via phase-locked disruption of its temporal coherence.
        Nature communications, 12(1), 363.

    Examples
    --------
    >>> f0 = 2
    >>> filt_BW = f0 / 2
    >>> N = 1000
    >>> sfreq = N / (2 * np.pi)
    >>> t = np.arange(-2 * np.pi, 0, 1 / sfreq)
    >>> X = np.cos(2 * np.pi * f0 * t - np.pi / 4)
    >>> l_freq = f0 - filt_BW / 2
    >>> h_freq = f0 + filt_BW / 2
    >>> Xf = echt(X, l_freq, h_freq, sfreq)
    """

    def __init__(self, l_freq, h_freq, sfreq, n_fft=None, filt_order=2):
        self.l_freq = l_freq
        self.h_freq = h_freq
        self.sfreq = sfreq
        self.n_fft = n_fft
        self.filt_order = filt_order

        # attributes
        self.h_ = None
        self.coef_ = None

    def fit(self, X, y=None):
        """Fit the ECHT transform to the input signal.

        Parameters
        ----------
        X : ndarray, shape=(n_samples, n_channels)
            The input signal to be transformed.

        """
        if self.n_fft is None:
            self.n_fft = next_fast_len(X.shape[0])

        # Set the amplitude of the negative components of the FFT to zero and
        # multiply the amplitudes of the positive components, apart from the
        # zero-frequency component (DC) and Nyquist components, by 2.
        #
        # If the signal has an even number of elements n, the frequency components
        # are:
        # - n/2-1 negative elements,
        # - one DC element,
        # - n/2-1 positive elements and
        # - one Nyquist element (in order).
        #
        # If the signal has an odd number of elements n, the frequency components
        # are:
        # - (n-1)/2 negative elements,
        # - one DC element,
        # - (n-1)/2 positive elements (in order).
        # - no positive element corresponding to the Nyquist frequency.
        self.h_ = np.zeros(self.n_fft)
        self.h_[0] = 1
        self.h_[1:(self.n_fft // 2) + 1] = 2
        if self.n_fft % 2 == 0:
            self.h_[self.n_fft // 2] = 1

        # The frequency response vector is computed using freqz from the filter's
        # impulse response function, computed by butter function, and the user
        # defined low-cutoff frequency, high-cutoff frequency and sampling rate.
        Wn = [self.l_freq / (self.sfreq / 2), self.h_freq / (self.sfreq / 2)]
        b, a = butter(self.filt_order, Wn, btype="band")
        T = 1 / self.sfreq * self.n_fft
        filt_freq = np.ceil(np.arange(-self.n_fft / 2, self.n_fft / 2) / T)

        self.coef_ = freqz(b, a, filt_freq, fs=self.sfreq)[1]
        self.coef_ = self.coef_[:, None]

        return self

    def transform(self, X):
        """Apply the ECHT transform to the input signal.

        Parameters
        ----------
        X : ndarray, shape=(n_samples, n_channels)
            The input signal to be transformed.

        Returns
        -------
        Xf : ndarray, shape=(n_samples, n_channels)
            The transformed signal (complex-valued).

        """
        if not np.isrealobj(X):
            X = np.real(X)

        # if not fitted
        if self.h_ is None or self.coef_ is None:
            self.fit(X)

        if X.ndim == 1:
            X = X[:, np.newaxis]

        # Compute the FFT of the signal.
        Xf = fft(X, self.n_fft, axis=0)

        # In contrast to :meth:`scipy.signal.hilbert()`, the code then
        # multiplies the array by a frequency response vector of a causal
        # bandpass filter.
        Xf = Xf * self.h_[:, None]

        # The array is arranged, using fft_shift function, so that the zero-frequency
        # component is at the center of the array, before the multiplication, and
        # rearranged back so that the zero-frequency component is at the left of the
        # array using ifft_shift(). Finally, the IFFT is computed.
        Xf = fftshift(Xf)
        Xf = Xf * self.coef_
        Xf = ifftshift(Xf)
        Xf = ifft(Xf, axis=0)

        return Xf

    def fit_transform(self, X, y=None):
        """Fit the ECHT transform to the input signal and transform it."""
        return self.fit(X).transform(X)


"""Test SNS functions."""
from numpy.testing import assert_allclose
from scipy.io import loadmat

from meegkit import sns


def test_sns():
    """Test against NoiseTools."""
    mat = loadmat("./tests/data/snsdata.mat")
    x = mat["x"]
    y_sns = mat["y_sns"]
    r_sns0 = mat["y_sns0"]
    cx = mat["cx"]

    r = sns.sns0(cx, n_neighbors=4)
    assert_allclose(r, r_sns0)  # assert our results match Matlab's

    y, _ = sns.sns(x, n_neighbors=4)
    assert_allclose(y, y_sns)  # assert our results match Matlab's


if __name__ == "__main__":
    import pytest
    pytest.main([__file__])



"""Test STAR functions."""
import numpy as np
from numpy.testing import assert_allclose

from meegkit.star import star
from meegkit.utils import demean, normcol

rng = np.random.default_rng(9)

def test_star1():
    """Test STAR 1."""
    # N channels,  1 sinusoidal target, N-3 noise sources, temporally local
    # artifacts on each channel.
    n_samples = 1000
    n_chans = 10
    f = 2

    def sim_data(n_samples, n_chans, f, SNR):
        target = np.sin(np.arange(n_samples) / n_samples * 2 * np.pi * f)
        target = target[:, np.newaxis]
        noise = rng.standard_normal((n_samples, n_chans - 3))

        x0 = normcol(np.dot(noise, rng.standard_normal((noise.shape[1], n_chans)))) \
            + SNR * target * rng.standard_normal((1, n_chans))
        x0 = demean(x0)
        artifact = np.zeros(x0.shape)
        for k in np.arange(n_chans):
            artifact[k * 100 + np.arange(20), k] = 1
        x = x0 + 20 * artifact
        return x, x0

    # Test SNR=1
    x, x0 = sim_data(n_samples, n_chans, f, SNR=np.sqrt(1))
    y, w, _ = star(x, 2, verbose="debug")
    assert_allclose(demean(y), x0)  # check that denoised signal ~ x0

    # Test more unfavourable SNR
    x, x0 = sim_data(n_samples, n_chans, f, SNR=np.sqrt(1e-7))
    y, w, _ = star(x, 2)
    assert_allclose(demean(y), x0)  # check that denoised signal ~ x0

    # Test an all-zero channel is properly handled
    x = x - x[:, 0][:, np.newaxis]
    y, w, _ = star(x, 2)
    assert_allclose(demean(y)[:, 0], x[:, 0])


if __name__ == "__main__":
    import pytest
    pytest.main([__file__])
    # test_star1()



"""Test bicoherence functions."""
import matplotlib.pyplot as plt
import numpy as np
import pytest
from scipy.fftpack import next_fast_len

from meegkit.utils.coherence import (
    cross_coherence,
    plot_polycoherence,
    plot_polycoherence_1d,
    plot_signal,
    polycoherence_0d,
    polycoherence_1d,
    polycoherence_1d_sum,
    polycoherence_2d,
)


@pytest.mark.parametrize("norm", [None, 2])
def test_coherence(norm, show=True):
    rng = np.random.default_rng(54326)

    # create signal with 5Hz and 8Hz components and noise, as well as a
    # 5Hz-8Hz interaction
    N = 10001
    kw = dict(nperseg=N // 4, noverlap=N // 20, nfft=next_fast_len(N // 2))
    t = np.linspace(0, 100, N)
    fs = 1 / (t[1] - t[0])
    s1 = np.cos(2 * np.pi * 5 * t + 0.3)  # 5Hz
    s2 = 3 * np.cos(2 * np.pi * 8 * t + 0.5)  # 8Hz
    noise = 5 * rng.normal(0, 1, N)
    signal = s1 + s2 + noise + 0.5 * s1 * s2

    # bicoherence
    # ---------------------------------------------r----------------------------

    # 0D local bicoherence (fixed frequencies)
    p5_8 = polycoherence_0d(signal, fs, [5, 8], **kw)
    p5_6 = polycoherence_0d(signal, fs, [5, 6], **kw)
    p5_6_8 = polycoherence_0d(signal, fs, [5, 6, 8], **kw)
    if norm is not None:
        print(f"bicoherence for f1=5Hz, f2=8Hz: {p5_8:.2f}")
        print(f"bicoherence for f1=5Hz, f2=6Hz: {p5_6:.2f}")
        print(f"bicoherence for f1=5Hz, f2=6Hz, f3=7Hz: {p5_6_8:.2f}")
        assert p5_8 > 0.85 > p5_6 > p5_6_8  # 5Hz and 7Hz are coherent, 5Hz and 6Hz not

    # 1D bicoherence (fixed f2)
    freqs, coh1d = polycoherence_1d(signal, fs, [5], **kw)

    # 1D bicoherence with sum (fixed f1+f2)
    freqs1dsum, coh1dsum = polycoherence_1d_sum(signal, fs, 13, **kw)

    # 2D bicoherence, span all frequencies
    # assert peaks at intersection of 5Hz and 8Hz, and 8-5=3Hz
    freqs1, freqs2, coh2d = polycoherence_2d(signal, fs, **kw)

    if norm is not None:
        assert np.max(coh2d) > 0.85
        assert np.abs(coh2d[freqs1 == 5, freqs2 == 8]) > 0.85
        assert np.abs(coh2d[freqs1 == 5, freqs2 == 3]) > 0.85


    if show:
        # Plot signal
        plot_signal(t, signal)
        plt.suptitle("signal and spectrum for bicoherence tests")

        # Plot bicoherence
        plot_polycoherence(freqs1, freqs2, coh2d)
        plt.suptitle("bicoherence")

        # Plot bicoherence 1D
        plot_polycoherence_1d(freqs, coh1d)
        plt.suptitle("bicoherence for f2=5Hz (column, expected 3Hz, 8Hz)")

        # Plot bicoherence for f1+f2=13Hz
        plot_polycoherence_1d(freqs1dsum, coh1dsum)
        plt.suptitle("bicoherence for f1+f2=13Hz (diagonal, expected 5Hz, 8Hz)")
        plt.show()


    # bicoherence with synthetic signal
    # -------------------------------------------------------------------------
    s3 = 4 * np.cos(2 * np.pi * 1 * t + 0.1)
    s5 = 0.4 * np.cos(2 * np.pi * 0.2 * t + 1)
    signal = s2 + s3 + s5 - 0.5 * s2 * s3 * s5 + noise

    synthetic = ((0.2, 10, 1), )
    p02_1_8 = polycoherence_0d(signal, fs, [0.2, 1, 8], synthetic=None,
                               norm=norm, **kw)
    p02_1_8s = polycoherence_0d(signal, fs, [0.2, 1, 8], synthetic=synthetic,
                               norm=norm, **kw)
    if norm is not None:
        print(f"coherence for f1=0.02Hz, f2=1Hz, f3=7Hz: {p02_1_8:.2f}")
        print(f"coherence for f1=0.02Hz (synthetic), f2=1Hz, f3=7Hz: {p02_1_8s:.2f}")
        assert p02_1_8s > 0.9 > p02_1_8

    result = polycoherence_2d(signal, fs, [0.02], synthetic=synthetic, norm=norm, **kw)

    if show:
        plot_signal(t, signal)
        plt.suptitle("signal and spectrum for tricoherence with synthetic signals")
        plt.tight_layout()

        plot_polycoherence(*result)
        plt.suptitle("tricoherence with f3=0.2Hz (synthetic)")
        plt.show()

    # cross-coherence
    # -------------------------------------------------------------------------
    s1 = np.cos(2 * np.pi * 5 * t + 0.3)  # 5Hz
    s2 = 3 * np.cos(2 * np.pi * 8 * t + 0.5)  # 8Hz
    signal1 = s1 + 4 * rng.normal(0, 1, N)
    signal2 = s2 + s1 + 5 * rng.normal(0, 1, N)
    f, Cxy = cross_coherence(signal1, signal2, fs, norm=None, **kw)

    if show:
        plot_polycoherence(f, f, Cxy)
        plt.suptitle("cross-coherence between 5Hz and 5+8Hz signals")
        plt.tight_layout()
        plt.show()

@pytest.mark.parametrize("shape", [(1001,), (3, 1001), (17, 3, 1001)])
def test_coherence_shapes(shape):
    """Test coherence functions with 1D, 2D or 3D input data."""
    rng = np.random.default_rng(54326)

    # create signal with 5Hz and 8Hz components and noise, as well as a
    # 5Hz-8Hz interaction
    N = shape[-1]
    t = np.linspace(0, 100, N)
    fs = 1 / (t[1] - t[0])
    s1 = np.cos(2 * np.pi * 5 * t + 0.3)  # 5Hz
    s2 = 3 * np.cos(2 * np.pi * 8 * t + 0.5)  # 8Hz
    noise = 5 * rng.normal(0, 1, shape)
    signal = np.broadcast_to(s1 + s2, shape) + noise

    assert signal.shape == shape

    B = polycoherence_0d(signal, fs, [5, 8])
    assert B.shape == shape[:-1]
    f, B = polycoherence_1d(signal, fs, [8])
    assert B.shape == shape[:-1] + f.shape
    f, B = polycoherence_1d_sum(signal, fs, 13)
    assert B.shape == shape[:-1] + f.shape
    f1, f2, B = polycoherence_2d(signal, fs)
    assert B.shape == shape[:-1] + f1.shape + f2.shape


if __name__ == "__main__":
    pytest.main([__file__])
    # test_coherence(2, False)


"""
Robust detrending examples
==========================

Some toy examples to showcase usage for ``meegkit.detrend`` module.

Robust referencing is adapted from [1].

References
----------
> [1] de Cheveigné, A., & Arzounian, D. (2018). Robust detrending,
  rereferencing, outlier detection, and inpainting for multichannel data.
  NeuroImage, 172, 903-912.

"""
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.gridspec import GridSpec

from meegkit.detrend import detrend, regress

# import config  # plotting utils

rng = np.random.default_rng(9)

###############################################################################
# Regression
# =============================================================================

###############################################################################
# Simple regression example, no weights
# -----------------------------------------------------------------------------
# We first try to fit a simple random walk process.
x = np.cumsum(rng.standard_normal((1000, 1)), axis=0)
r = np.arange(1000.)[:, None]
r = np.hstack([r, r ** 2, r ** 3])
b, y = regress(x, r)

plt.figure(1)
plt.plot(x, label="data")
plt.plot(y, label="fit")
plt.title("No weights")
plt.legend()
plt.show()

###############################################################################
# Downweight 1st half of the data
# -----------------------------------------------------------------------------
# We can also use weights for each time sample. Here we explicitly restrict the
# fit to the second half of the data by setting weights to zero for the first
# 500 samples.
x = np.cumsum(rng.standard_normal((1000, 1)), axis=0) + 1000
w = np.ones(y.shape[0])
w[:500] = 0
b, y = regress(x, r, w)

f = plt.figure(3)
gs = GridSpec(4, 1, figure=f)
ax1 = f.add_subplot(gs[:3, 0])
ax1.plot(x, label="data")
ax1.plot(y, label="fit")
ax1.set_xticklabels("")
ax1.set_title("Split-wise regression")
ax1.legend()
ax2 = f.add_subplot(gs[3, 0])
ll, = ax2.plot(np.arange(1000), np.zeros(1000))
ax2.stackplot(np.arange(1000), w, labels=["weights"], color=ll.get_color())
ax2.legend(loc=2)

###############################################################################
# Multichannel regression
# -----------------------------------------------------------------------------
x = np.cumsum(rng.standard_normal((1000, 2)), axis=0)
w = np.ones(y.shape[0])
b, y = regress(x, r, w)

plt.figure(4)
plt.plot(x, label="data", color="C0")
plt.plot(y, ls=":", label="fit", color="C1")
plt.title("Channel-wise regression")
plt.legend()


###############################################################################
# Detrending
# =============================================================================

###############################################################################
# Basic example with a linear trend
# -----------------------------------------------------------------------------
x = np.arange(100)[:, None]
x = x + rng.standard_normal(x.shape)
y, _, _ = detrend(x, 1)

plt.figure(5)
plt.plot(x, label="original")
plt.plot(y, label="detrended")
plt.legend()

###############################################################################
# Detrend biased random walk with a third-order polynomial
# -----------------------------------------------------------------------------
x = np.cumsum(rng.standard_normal((1000, 1)) + 0.1)
y, _, _ = detrend(x, 3)

plt.figure(6)
plt.plot(x, label="original")
plt.plot(y, label="detrended")
plt.legend()

###############################################################################
# Detrend with weights
# -----------------------------------------------------------------------------
# Finally, we show how the detrending process handles local artifacts, and how
# we can advantageously use weights to improve detrending. The raw data
# consists of gaussian noise with a linear trend, and a storng glitch covering
# the first 100 timesamples (blue trace). Detrending without weights (orange
# trace) causes an overestimation of the polynomial order because of the
# glitch, leading to a mediocre fit. When downweightining this artifactual
# period, the fit is much improved (green trace).
x = np.linspace(0, 100, 1000)[:, None]
x = x + 3 * rng.standard_normal(x.shape)

# introduce some strong artifact on the first 100 samples
x[:100, :] = 100

# Detrend
y, _, _ = detrend(x, 3, None, threshold=np.inf)

# Same process but this time downweight artifactual window
w = np.ones(x.shape)
w[:100, :] = 0
z, _, _ = detrend(x, 3, w)

plt.figure(7)
plt.plot(x, label="original")
plt.plot(y, label="detrended - no weights")
plt.plot(z, label="detrended - weights")
plt.legend()
plt.show()



"""
ASR example
===========

Denoise data using Artifact Subspace Reconstruction.

Uses meegkit.ASR().
"""
import os

import matplotlib.pyplot as plt
import numpy as np

from meegkit.asr import ASR
from meegkit.utils.matrix import sliding_window

# THIS_FOLDER = os.path.dirname(os.path.abspath(__file__))
raw = np.load(os.path.join("..", "tests", "data", "eeg_raw.npy"))
sfreq = 250

###############################################################################
# Calibration and processing
# -----------------------------------------------------------------------------

# Train on a clean portion of data
asr = ASR(method="euclid")
train_idx = np.arange(0 * sfreq, 30 * sfreq, dtype=int)
_, sample_mask = asr.fit(raw[:, train_idx])

# Apply filter using sliding (non-overlapping) windows
X = sliding_window(raw, window=int(sfreq), step=int(sfreq))
Y = np.zeros_like(X)
for i in range(X.shape[1]):
    Y[:, i, :] = asr.transform(X[:, i, :])

raw = X.reshape(8, -1)  # reshape to (n_chans, n_times)
clean = Y.reshape(8, -1)

###############################################################################
# Plot the results
# -----------------------------------------------------------------------------
#
# Data was trained on a 40s window from 5s to 45s onwards (gray filled area).
# The algorithm then removes portions of this data with high amplitude
# artifacts before running the calibration (hatched area = good).

times = np.arange(raw.shape[-1]) / sfreq
f, ax = plt.subplots(8, sharex=True, figsize=(8, 5))
for i in range(8):
    ax[i].fill_between(train_idx / sfreq, 0, 1, color="grey", alpha=.3,
                       transform=ax[i].get_xaxis_transform(),
                       label="calibration window")
    ax[i].fill_between(train_idx / sfreq, 0, 1, where=sample_mask.flat,
                       transform=ax[i].get_xaxis_transform(),
                       facecolor="none", hatch="...", edgecolor="k",
                       label="selected window")
    ax[i].plot(times, raw[i], lw=.5, label="before ASR")
    ax[i].plot(times, clean[i], label="after ASR", lw=.5)
    ax[i].set_ylim([-50, 50])
    ax[i].set_ylabel(f"ch{i}")
    ax[i].set_yticks([])
ax[i].set_xlabel("Time (s)")
ax[0].legend(fontsize="small", bbox_to_anchor=(1.04, 1), borderaxespad=0)
plt.subplots_adjust(hspace=0, right=0.75)
plt.suptitle("Before/after ASR")
plt.show()



"""
Example demonstrating STAR + DSS
================================

This example shows how one can effectively combine STAR and DSS to recover
signal components which would not have been discoverable with either these
two techniques alone, due to the presence of strong artifacts.

This example replicates figure 1 in [1]_.

References
----------
.. [1] de Cheveigné A (2016) Sparse Time Artifact Removal, Journal of
   Neuroscience Methods, 262, 14-20, doi:10.1016/j.jneumeth.2016.01.005

"""
import matplotlib.pyplot as plt
import numpy as np
from scipy.optimize import leastsq

from meegkit import dss, star
from meegkit.utils import demean, normcol, tscov

# import config  # noqa

rng = np.random.default_rng(9)

###############################################################################
# Create simulated data
# -----------------------------------------------------------------------------
# Simulated data consist of N channels, 1 sinusoidal target, N-3 noise sources,
# with temporally local artifacts on each channel.

# Source
n_chans, n_samples = 10, 1000
f = 2
target = np.sin(np.arange(n_samples) / n_samples * 2 * np.pi * f)
target = target[:, np.newaxis]
noise = rng.standard_normal((n_samples, n_chans - 3))

# Create artifact signal
SNR = np.sqrt(1)
x0 = normcol(np.dot(noise, rng.standard_normal((noise.shape[1], n_chans)))) + \
    SNR * target * rng.standard_normal((1, n_chans))
x0 = demean(x0)
artifact = np.zeros(x0.shape)
for k in np.arange(n_chans):
    artifact[k * 100 + np.arange(20), k] = 1
x = x0 + 10 * artifact


def _sine_fit(x):
    """Fit a sinusoidal trend."""
    guess_mean = np.mean(x)
    guess_std = np.std(x)
    guess_phase = 0
    t = np.linspace(0, 4 * np.pi, x.shape[0])

    # Optimization function, in this case, we want to minimize the difference
    # between the actual data and our "guessed" parameters
    def func(y):
        return np.mean(x - (y[0] * np.sin(t + y[1]) + y[2])[:, None], 1)

    est_std, est_phase, est_mean = leastsq(
        func, [guess_std, guess_phase, guess_mean])[0]
    data_fit = est_std * np.sin(t + est_phase) + est_mean
    return np.tile(data_fit, (x.shape[1], 1)).T


###############################################################################
# 1) Apply STAR
# -----------------------------------------------------------------------------
y, w, _ = star.star(x, 2)

###############################################################################
# 2) Apply DSS on raw data
# -----------------------------------------------------------------------------
c0, _ = tscov(x)
c1, _ = tscov(x - _sine_fit(x))
[todss, _, pwr0, pwr1] = dss.dss0(c0, c1)
z1 = normcol(np.dot(x, todss))

###############################################################################
# 3) Apply DSS on STAR-ed data
# -----------------------------------------------------------------------------
# Here the bias function is the original signal minus the sinusoidal trend.
c0, _ = tscov(y)
c1, _ = tscov(y - _sine_fit(y))
[todss, _, pwr0, pwr1] = dss.dss0(c0, c1)
z2 = normcol(np.dot(y, todss))

###############################################################################
# Plots
# -----------------------------------------------------------------------------
f, (ax0, ax1, ax2, ax3) = plt.subplots(4, 1, figsize=(7, 9))
ax0.plot(target, lw=.5)
ax0.set_title("Target")

ax1.plot(x, lw=.5)
ax1.set_title(f"Signal + Artifacts (SNR = {SNR})")

ax2.plot(z1[:, 0], lw=.5, label="Best DSS component")
ax2.set_title("DSS")
ax2.legend(loc="lower right")

ax3.plot(z2[:, 0], lw=.5, label="Best DSS component")
ax3.set_title("STAR + DSS")
ax3.legend(loc="lower right")

f.set_tight_layout(True)
plt.show()



"""Covariance calculation."""
import numpy as np
from scipy import linalg

from .base import mldivide
from .matrix import (
    _check_shifts,
    _check_weights,
    multishift,
    relshift,
    theshapeof,
    unsqueeze,
)

rng = np.random.default_rng()


def block_covariance(data, window=128, overlap=0.5, padding=True, estimator="cov"):
    """Compute blockwise covariance.

    Parameters
    ----------
    data : array, shape=(n_chans, n_samples)
        Input data (must be 2D)
    window : int
        Window size.
    overlap : float
        Overlap between successive windows.

    Returns
    -------
    cov : array, shape=(n_blocks, n_chans, n_chans)
        Block covariance.

    """
    from pyriemann.utils.covariance import check_function, cov_est_functions

    assert 0 <= overlap < 1, "overlap must be < 1"
    est = check_function(estimator, cov_est_functions)
    cov = []
    n_chans, n_samples = data.shape
    if padding:  # pad data with zeros
        pad = np.zeros((n_chans, int(window / 2)))
        data = np.concatenate((pad, data, pad), axis=1)

    jump = int(window * overlap)
    ix = 0
    while (ix + window < n_samples):
        cov.append(est(data[:, ix:ix + window]))
        ix = ix + jump

    return np.array(cov)


def cov_lags(X, Y, shifts=None):
    """Empirical covariance of the joint array [X, Y] with lags.

    Parameters
    ----------
    X: array, shape=(n_times, n_chans_x[, n_trials])
        Time shifted data.
    Y: array, shape=(n_times, n_chans_y[, n_trials])
        Reference data.
    shifts: array, shape=(n_shifts,)
        Positive lag means X is delayed relative to Y.

    Returns
    -------
    C : array, shape=(n_chans_x + n_chans_y, n_chans_x + n_chans_y, n_shifts)
        Covariance matrix (3D if n_shifts > 1).
    tw : float
        Total weight.
    m : int
        Number of columns in X.

    See Also
    --------
    relshift, tscov, tsxcov

    """
    shifts, n_shifts = _check_shifts(shifts)
    X, Y = unsqueeze(X), unsqueeze(Y)

    n_samples, n_chans, n_trials = theshapeof(X)
    n_samples2, n_chans2, n_trials2 = theshapeof(Y)

    if n_samples != n_samples2:
        raise AttributeError("X and Y must have same n_times")
    if n_trials != n_trials2:
        raise AttributeError("X and Y must have same n_trials")
    if n_samples <= max(shifts):
        raise AttributeError("shifts should be no larger than n_samples")

    n_cov = n_chans + n_chans2  # sum of channels of X and Y
    C = np.zeros((n_cov, n_cov, n_shifts))
    for t in np.arange(n_trials):
        for i, s in enumerate(shifts):
            XX, YY = relshift(X[..., t], ref=Y[..., t], shifts=s)
            XY = np.hstack((XX, YY))
            C[:, :, i] += np.dot(XY.T, XY)

    if n_shifts == 1:
        C = np.squeeze(C, 2)

    tw = n_samples * n_trials

    return C, tw, n_chans


def tsxcov(X, Y, shifts=None, weights=None, assume_centered=True):
    """Calculate cross-covariance of X and time-shifted Y.

    This function calculates, for each pair of columns (Xi, Yj) of X and Y, the
    scalar products between Xi and time-shifted versions of Yj.

    Output is a 2D matrix with dimensions .

    Parameters
    ----------
    X, Y : arrays, shape=(n_times, n_chans[, n_trials])
        Data to cross correlate. X can be 1D, 2D or 3D.
    shifts : array
        Time shifts.
    weights : array
        The weights that are applied to X. 1D (if X is 1D or 2D) or 2D (if X is
        3D).
    assume_centered : bool
        If False, remove the mean of X before computing the covariance
        (default=True).

    Returns
    -------
    C : array, shape=(n_chans_x, n_chans_y * n_shifts)
        Cross-covariance matrix.
    tw : total weight

    """
    n_times, n_chans, n_trials = theshapeof(X)
    n_times2, n_chans2, n_trials2 = theshapeof(Y)
    X = unsqueeze(X)
    Y = unsqueeze(Y)

    weights = _check_weights(weights, X)
    shifts, n_shifts = _check_shifts(shifts)

    if not assume_centered:
        X = X - X.mean(0, keepdims=1)
        Y = Y - Y.mean(0, keepdims=1)

    # Apply weights if any
    if weights.any():
        X = np.einsum("ijk,ilk->ijk", X, weights)  # element-wise mult
        weights = weights[:n_times2, :, :]

    # cross covariance
    # C = np.zeros((n_chans * n_shifts, n_chans2 * n_shifts))
    # for t in np.arange(n_trials):
    #     YY, XX = relshift(Y[..., t], ref=X[..., t], shifts=shifts)
    #     XX = XX.reshape(n_times, n_chans * n_shifts)
    #     YY = YY.reshape(n_times2, n_chans2 * n_shifts)
    #     C += np.dot(XX.T, YY)
    C = np.zeros((n_chans, n_chans2 * n_shifts))
    for t in np.arange(n_trials):
        YY = multishift(Y[..., t], shifts=shifts)
        YY = YY.reshape(n_times2, n_chans2 * n_shifts)
        C += np.dot(X[..., t].T, YY)

    if not weights.any():
        tw = n_trials * n_chans2 * YY.shape[0]
    else:
        weights = weights[:YY.shape[0], ...]
        tw = np.sum(weights.flat)

    return C, tw


def tscov(X, shifts=None, weights=None, assume_centered=True):
    """Time shift covariance.

    This function calculates, for each pair [X[i], X[j]] of columns of X, the
    cross-covariance matrix between the time-shifted versions of X[i].

    Parameters
    ----------
    X : array, shape=(n_times, n_chans[, n_trials])
        Data, can be 1D, 2D or 3D.
    shifts : array, shape=(n_shifts,) | None
        Array of time shifts.
    weights : array, shape=(n_times,) or
            shape=(n_times, 1, n_trials), or
            shape=(n_times, n_chans, n_trials)

        Weights, 1D (if X is 1D or 2D) or 2D (if X is 3D). The weights are not
        shifted.
    assume_centered : bool
        If False, remove the mean of X before computing the covariance
        (default=True).

    Returns
    -------
    C : array, shape=(n_chans * n_shifts, n_chans * n_shifts)
        Covariance matrix. This matrix is made up of a (n_times, n_times)
        matrix of submatrices of dimensions (n_shifts, n_shifts).
    tw : array
        Total weight (C/tw is the normalized covariance).

    """
    n_times, n_chans, n_trials = theshapeof(X)
    X = unsqueeze(X)

    weights = _check_weights(weights, X)
    shifts, n_shifts = _check_shifts(shifts)

    if not assume_centered:
        X = X - X.mean(0, keepdims=True)

    if weights.any():  # weights
        X = np.einsum("ijk,ilk->ijk", X, weights)  # element-wise mult
        tw = np.sum(weights[:]) - 1
    else:  # no weights
        N = 0
        if len(shifts[shifts < 0]):
            N -= np.min(shifts)
        if len(shifts[shifts >= 0]):
            N += np.max(shifts)
        tw = n_trials * (n_times - N - 1)

    C = np.zeros((n_chans * n_shifts, n_chans * n_shifts))
    for trial in range(n_trials):
        XX = multishift(X[..., trial], shifts)
        XX = XX.reshape(n_times, n_chans * n_shifts)
        C += np.dot(XX.T, XX)

    return C, tw


def convmtx(V, n):
    """Generate a convolution matrix.

    CONVMTX(V,N) returns the convolution matrix for vector V. If V is a column
    vector and X is a column vector of length N, then CONVMTX(V,N)*X is the
    same as CONV(V,X). If R is a row vector and X is a row vector of length N,
    then X*CONVMTX(R,N) is the same as CONV(R,X).

    Given a vector V of length N, an N+n-1 by n convolution matrix is
    generated of the following form:

    ::

            |  V(0)  0      0     ...      0    |
            |  V(1) V(0)    0     ...      0    |
            |  V(2) V(1)   V(0)   ...      0    |
        X = |   .    .      .              .    |
            |   .    .      .              .    |
            |   .    .      .              .    |
            |  V(N) V(N-1) V(N-2) ...  V(N-n+1) |
            |   0   V(N)   V(N-1) ...  V(N-n+2) |
            |   .    .      .              .    |
            |   .    .      .              .    |
            |   0    0      0     ...    V(N)   |

    That is, V is assumed to be causal, and zero-valued after N.

    Parameters
    ----------
    V : array, shape=(N,) or(N, 1) or (1, N)
        Input vector.
    n : int

    Returns
    -------
    t : array, shape=(N * n - 1, n)

    Examples
    --------
    Generate a simple convolution matrix:

    >>> h = [1, 2, 1]
    >>> convmtx(h,7)
    np.array(
        [[1. 2. 1. 0. 0. 0.]
         [0. 1. 2. 1. 0. 0.]
         [0. 0. 1. 2. 1. 0.]
         [0. 0. 0. 1. 2. 1.]]
    )

    """
    V = np.asarray(V)
    if V.ndim == 1:
        V = V[:, None]
    else:
        assert V.shape[0] == 1 or V.shape[1] == 1

    [nr, nc] = V.shape
    V = V.flatten()

    c = np.hstack((V, np.zeros(n - 1)))
    r = np.zeros(n)
    m = len(c)
    x_left = r[n:0:-1]  # reverse order from n to 2 in original code
    x_right = c.flatten()
    x = np.hstack((x_left, x_right))
    cidx = np.arange(0., (m - 1.) + 1).conj().T
    ridx = np.arange(n, (1.) + (-1.), -1.)

    t = np.zeros([len(cidx), len(ridx)])
    counter_cidx = 0
    for c_val in cidx:
        counter_ridx = 0
        for r_val in ridx:
            t[counter_cidx, counter_ridx] = c_val + r_val
            counter_ridx += 1
        counter_cidx += 1

    # Toeplitz subscripts
    t[:] = x[t.astype(int) - 1]

    if nr > nc:
        t = t.T

    return t


def pca(cov, max_comps=None, thresh=0):
    """PCA from covariance.

    Parameters
    ----------
    cov:  array, shape=(n_chans, n_chans)
        Covariance matrix.
    max_comps : int | None
        Maximum number of components to retain after decomposition. ``None``
        (the default) keeps all suprathreshold components (see ``thresh``).
    thresh : float
        Discard components below this threshold.

    Returns
    -------
    V : array, shape=(max_comps, max_comps)
        Eigenvectors (matrix of PCA components).
    d : array, shape=(max_comps,)
        PCA eigenvalues

    """
    if thresh is not None and (thresh > 1 or thresh < 0):
        raise ValueError("Threshold must be between 0 and 1 (or None).")

    d, V = linalg.eigh(cov)
    d = d.real
    V = V.real

    p0 = d.sum()  # total power

    idx = np.argsort(d)[::-1]  # reverse sort ev order
    d = d[idx]
    V = V[:, idx]

    # Truncate weak components
    if thresh is not None:
        idx = np.where(d / d.max() > thresh)[0]
        d = d[idx]
        V = V[:, idx]

    # Keep a fixed number of components
    if max_comps is None:
        max_comps = V.shape[1]
    else:
        max_comps = np.min((max_comps, V.shape[1]))

    V = V[:, np.arange(max_comps)]
    d = d[np.arange(max_comps)]

    var = 100 * d.sum() / p0
    if var < 99:
        print(f"[PCA] Explained variance of selected components : {var:.2f}%")

    return V, d


def regcov(Cxy, Cyy, keep=None, threshold=0):
    """Compute regression matrix from cross covariance.

    Parameters
    ----------
    Cxy : array
        Cross-covariance matrix between data and regressor.
    Cyy : array
        Covariance matrix of regressor.
    keep : array
        Number of regressor PCs to keep (default=None, which keeps all).
    threshold : float
        Eigenvalue threshold for discarding regressor PCs (default=0, which keeps all
        components).

    Returns
    -------
    R : array
        Matrix to apply to regressor to best model data.

    """
    # PCA of regressor
    [V, d] = pca(Cyy, max_comps=keep, thresh=threshold)

    # cross-covariance between data and regressor PCs
    Cxy = Cxy.T
    R = np.dot(V.T, Cxy)

    # projection matrix from regressor PCs
    R = (R.T * 1 / d).T

    # projection matrix from regressors
    R = V @ R  # np.dot(np.squeeze(V), np.squeeze(R))

    # if R.ndim == 1:
    #     R = R[:, None]

    return R


def nonlinear_eigenspace(L, k, alpha=1):
    """Nonlinear eigenvalue problem: total energy minimization.

    This example is motivated in [1]_ and was adapted from the manopt toolbox
    in Matlab.

    TODO : check this

    Parameters
    ----------
    L : array, shape=(n_channels, n_channels)
        Discrete Laplacian operator: the covariance matrix.
    alpha : float
        Given constant for optimization problem.
    k : int
        Determines how many eigenvalues are returned.

    Returns
    -------
    Xsol : array, shape=(n_channels, n_channels)
        Eigenvectors.
    S0 : array
        Eigenvalues.

    References
    ----------
    .. [1] "A Riemannian Newton Algorithm for Nonlinear Eigenvalue Problems",
       Zhi Zhao, Zheng-Jian Bai, and Xiao-Qing Jin, SIAM Journal on Matrix
       Analysis and Applications, 36(2), 752-774, 2015.

    """
    import pymanopt
    from pymanopt import Problem
    from pymanopt.manifolds import Grassmann
    from pymanopt.optimizers import TrustRegions

    n = L.shape[0]
    assert L.shape[1] == n, "L must be square."

    # Grassmann manifold description
    manifold = Grassmann(n, k)
    manifold._dimension = 1  # hack

    # A solver that involves the hessian (check if correct TODO)
    solver = TrustRegions(verbosity=0)

    # Cost function evaluation
    @pymanopt.function.numpy(manifold)
    def cost(X):
        rhoX = np.sum(X ** 2, 1, keepdims=True)  # diag(X*X')
        val = 0.5 * np.trace(X.T @ (L * X)) + \
            (alpha / 4) * (rhoX.T @ mldivide(L, rhoX))
        return val

    # Euclidean gradient evaluation
    @pymanopt.function.numpy(manifold)
    def egrad(X):
        rhoX = np.sum(X ** 2, 1, keepdims=True)  # diag(X*X')
        g = L @ X + alpha * np.diagflat(mldivide(L, rhoX)) @ X
        return g

    # Euclidean Hessian evaluation
    # Note: Manopt automatically converts it to the Riemannian counterpart.
    @pymanopt.function.numpy(manifold)
    def ehess(X, U):
        rhoX = np.sum(X ** 2, 1, keepdims=True)  # np.diag(X * X')
        rhoXdot = 2 * np.sum(X.dot(U), 1)
        h = L @ U + alpha * np.diagflat(mldivide(L, rhoXdot)) @ X + \
            alpha * np.diagflat(mldivide(L, rhoX)) @ U
        return h

    # Initialization as suggested in above referenced paper.
    # randomly generate starting point for svd
    x = rng.standard_normal((n, k))
    [U, S, V] = linalg.svd(x, full_matrices=False)
    x = U.dot(V.T)
    S0, U0 = linalg.eig(
        L + alpha * np.diagflat(mldivide(L, np.sum(x**2, 1)))
    )

    # Call manoptsolve to automatically call an appropriate solver.
    # Note: it calls the trust regions solver as we have all the required
    # ingredients, namely, gradient and Hessian, information.
    problem = Problem(manifold=manifold, cost=cost, euclidean_gradient=egrad,
                      euclidean_hessian=ehess)
    Xsol = solver.run(problem, initial_point=U0)

    return S0, Xsol.point



"""Buffer class for real-time processing."""
import warnings

import numpy as np


class Buffer:
    """Circular buffer for real-time processing.

    Parameters
    ----------
    size : int
        The number of samples of the buffer.
    n_channels : int
        The number of channels of the buffer.

    Attributes
    ----------
    size : int
        The number of samples of the buffer.
    n_channels : int
        The number of channels of the buffer.
    counter : int
        The number of samples in the buffer.
    _data : ndarray, shape (size, n_channels)
        Data buffer.

    """

    def __init__(self, size, n_channels=1):
        self.size = size
        self.n_channels = n_channels
        self.reset()

    def reset(self):
        """Reset the buffer."""
        self.counter = 0  # number of samples in the buffer
        self._head = 0  # most recent sample in the buffer
        self._tail = 0  # most recent read sample in the buffer
        self._data = np.zeros((self.size, self.n_channels))

    def push(self, X):
        """Update the buffer with new data.

        Parameters
        ----------
        X : ndarray, shape (n_samples, n_channels)
            The input signal to be transformed.
        """
        # Check input
        if X.ndim == 0:
            # X = np.array([X])
            n_samples = 1
        elif X.ndim == 1:
            X = X[:, None]
            n_samples = X.shape[0]
        elif X.ndim > 2:
            raise ValueError("X must be 1- or 2-dimensional")

        # Check size
        if n_samples > self.size:
            warnings.warn("Buffer overflow: some old data has been lost")
            X = X[-self.size:]  # keep only the last self.size samples
            n_samples = X.shape[0]

        # Warn when data has not been read for a long time
        if self.counter + n_samples - self._tail > self.size:
            warnings.warn("Buffer overflow: some old data has been discarded")
            self._tail = self._head + n_samples - self.size

        # Update buffer
        end = self.head + n_samples
        if end <= self.size:
            # If the new data fits in the remaining space
            self._data[self.head:end] = X
        else:
            # If the new data wraps around to the beginning of the buffer
            overflow = end - self.size
            self._data[self.head:] = X[:-overflow]
            self._data[:overflow] = X[-overflow:]

        self._head += n_samples
        self.counter += n_samples

        return self

    def get_new_samples(self, n_samples=None):
        """Consume n_samples."""
        if n_samples is None:
            n_samples = min(self.counter, self.size)
        elif n_samples > self.counter:
            raise ValueError("Not enough samples in the buffer")
        elif n_samples > self.size:
            raise ValueError("Requested more samples than buffer size")

        start = self.tail
        end = (self.tail + n_samples) % self.size

        if start < end:
            samples = self._data[start:end]
        else:
            samples = np.concatenate((self._data[start:], self._data[:end]))

        self._tail += n_samples
        return samples

    def view(self, n_samples):
        """View the last n_samples, without consuming them."""
        start = (self.head - n_samples) % self.size
        end = self.head
        if start < end:
            return self._data[start:end]
        else:
            return np.concatenate((self._data[start:], self._data[:end]))

    def __repr__(self) -> str:
        """Return complete string representation."""
        repr = f"Buffer({self.size}, {self.n_channels})\n"
        repr += f"> counter: {self.counter}\n"
        repr += f"> head: {self.head}\n"
        repr += f"> tail: {self.tail}\n"
        return repr

    @property
    def tail(self):
        """Get the index of the tail."""
        return self._tail % self.size

    @property
    def head(self):
        """Get the index of the head."""
        return self._head % self.size
